from lifeblood.basenode import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.enums import NodeParameterType
from lifeblood_alicevision_modules.base_node import AlicevisionBaseNode
from pathlib import Path
from typing import Iterable


def node_class():
    return AlicevisionMeshFiltering


class AlicevisionMeshFiltering(AlicevisionBaseNode):
    def __init__(self, name):
        super(AlicevisionMeshFiltering, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('inputMesh', 'Mesh', NodeParameterType.STRING, "`task['av_mesh_object']`")
            ui.add_parameter('outputMesh', 'Output Filtered Mesh', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/MeshFiltering/mesh.obj")
            ui.add_separator()
            ui.add_parameter('keepLargestMeshOnly', 'Keep Only the Largest Mesh', NodeParameterType.BOOL, False)
            ui.add_parameter('smoothingIterations', 'Smoothing Iterations', NodeParameterType.INT, 5)
            ui.add_parameter('filterLargeTrianglesFactor', 'Filter Large Triangles Factor', NodeParameterType.FLOAT, 60.0)

    @classmethod
    def label(cls) -> str:
        return 'AV Mesh Filtering'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'alicevision', 'meshing', 'filter'

    @classmethod
    def type_name(cls) -> str:
        return 'stock_alicevision_mesh_filtering'

    @classmethod
    def description(cls) -> str:
        return 'this node type does not have a description'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        output = Path(context.param_value('outputMesh').strip())
        if not output.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')

        output.parent.mkdir(parents=True, exist_ok=True)

        args = ['aliceVision_meshFiltering', '--verboseLevel', 'info']
        args += ['--inputMesh', context.param_value('inputMesh').strip()]
        for param_name in ('keepLargestMeshOnly', 'smoothingIterations', 'filterLargeTrianglesFactor'):
            args += [f'--{param_name}', context.param_value(param_name)]

        args += ['--outputMesh', output]

        job = InvocationJob([str(x) for x in args])
        res = ProcessingResult(job)
        res.set_attribute('av_mesh_object_filtered', str(output))
        return res



import os
import asyncio
import random
import uuid
import time
import psutil
from time import perf_counter
from contextlib import contextmanager, asynccontextmanager
from .logging import get_logger, logging

from typing import List, Optional, Union


class DummyLock:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass


def atimeit(threshold=0):
    def _atimeit(func):
        logger = get_logger('timeit')
        if not logger.isEnabledFor(logging.DEBUG):
            return func

        async def _wrapper(*args, **kwargs):
            _start = time.perf_counter()
            raised = False
            try:
                return await func(*args, **kwargs)
            except Exception:
                raised = True
                raise
            finally:
                dt = time.perf_counter() - _start
                if dt > threshold:
                    logger.debug(f'ran{"(raised)" if raised else ""} {func.__name__} in {dt}s')

        return _wrapper

    return _atimeit


def timeit(threshold=0):
    def _timeit(func):
        logger = get_logger('timeit')
        if not logger.isEnabledFor(logging.DEBUG):
            return func

        def _wrapper(*args, **kwargs):
            _start = time.perf_counter()
            raised = False
            try:
                return func(*args, **kwargs)
            except Exception:
                raised = True
                raise
            finally:
                dt = time.perf_counter() - _start
                if dt > threshold:
                    logger.debug(f'ran{"(raised)" if raised else ""} {func.__name__} in {dt}s')
        return _wrapper

    return _timeit


__alocking_locks = {}


def alocking(lock_or_name: Union[asyncio.Lock, str, None] = None):  # TODO: TESTS NEEDED!
    def decorator(func):
        async def _wrapper(*args, **kwargs):
            nonlocal lock_or_name
            lock = None
            if isinstance(lock_or_name, asyncio.Lock):
                lock = lock_or_name
            elif lock_or_name is None:
                lock_or_name = func.__name__

            if isinstance(lock_or_name, str):
                if lock_or_name not in __alocking_locks:
                    __alocking_locks[lock_or_name] = asyncio.Lock()
                lock = __alocking_locks[lock_or_name]

            if lock is None:
                raise ValueError('lock value bad')

            async with lock:
                return await func(*args, **kwargs)
        return _wrapper
    return decorator


__stashed_machine_uuid = None


def get_unique_machine_id() -> int:
    """
    get a unique machine id (or something as close as possible)

    To work well with sqlite, id must be truncated to 64 bit

    :return: 64 bit integer
    """
    global __stashed_machine_uuid
    logger = get_logger('get_unique_machine_id')

    if __stashed_machine_uuid is None:
        midpaths = ['/etc/machine-id', '/var/lib/dbus/machine-id']
        for midpath in midpaths:
            if os.path.exists(midpath) and os.access(midpath, os.R_OK):
                with open(midpath, 'r') as f:
                    __stashed_machine_uuid = int(f.read(), base=16)
                break
        else:
            try:
                ia_pairs = sorted([(i, a) for i, a in psutil.net_if_addrs().items()], key=lambda x: x[0])  # to be more predictable in order
                for _, addrlist in ia_pairs:
                    for addr in addrlist:
                        if addr.family != psutil.AF_LINK or addr.address == '00:00:00:00:00:00':
                            continue
                        address = addr.address
                        logger.debug(f'trying address: "{address}"')
                        __stashed_machine_uuid = int(address.replace(':', '').replace('-', ''), base=16)
                        break
                    if __stashed_machine_uuid is not None:
                        break
            except Exception as e:
                logger.warning(f'failed to get any MAC address, because of {repr(e)}: {str(e)}')
            if __stashed_machine_uuid is None:
                __stashed_machine_uuid = uuid.getnode()  # this is not very reliable.

        __stashed_machine_uuid &= (1 << 63) - 1  # sqlite eats signed 64b integers, so we leave single bit. though we could still cast and use it...
    return __stashed_machine_uuid


class _TimeMeasurer:
    def __init__(self):
        self.__start = None
        self.__stop = None

    def start(self):
        self.__start = perf_counter()

    def stop(self):
        self.__stop = perf_counter()

    def elapsed(self):
        return self.__stop - self.__start


@contextmanager
def performance_measurer(threshold_to_report=None, name=None):
    tm = _TimeMeasurer()
    tm.start()
    try:
        yield tm
    finally:
        tm.stop()
        if threshold_to_report and tm.elapsed() > threshold_to_report:
            logging.debug(f'ran {name} in {tm.elapsed()}s')


@asynccontextmanager
async def aperformance_measurer(threshold_to_report=None, name=None):
    tm = _TimeMeasurer()
    tm.start()
    try:
        yield tm
    finally:
        tm.stop()
        if threshold_to_report and tm.elapsed() > threshold_to_report:
            logging.debug(f'ran {name} in {tm.elapsed()}s')



import sys
import argparse

from . import simple_worker_pool


def main(argv):
    parser = argparse.ArgumentParser('lifeblood pool')
    mut = parser.add_mutually_exclusive_group()
    mut.add_argument('--list', action='store_true', help='list availabele pool types')
    mut.add_argument('type', nargs='?', help='worker pool type to start')

    # a bit cheaty, but here we always have fixed amount of args
    opts = parser.parse_args(argv[:1])
    remaining_args = argv[1:]

    known_types = {'simple': simple_worker_pool}
    if opts.list:
        print('known pool types:\n' + '\n'.join(f'\t{x}' for x in known_types))
        return

    # for now it's hardcoded logic, in future TODO: make worker pools extendable as plugins
    if opts.type is None:
        print('no pool type provided!')
        parser.print_help()
        return 2
    if opts.type not in known_types:
        print(f'unknown pool type "{opts.type}"\nuse --list flag to list known pool types', file=sys.stderr)
        return 2

    known_types[opts.type].main(remaining_args)



import json
from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult
from lifeblood.taskspawn import TaskSpawn
from lifeblood.exceptions import NodeNotReadyToProcess
from lifeblood.enums import NodeParameterType
from lifeblood.uidata import NodeUi
from lifeblood.processingcontext import ProcessingContext

from threading import Lock

from typing import Dict, TypedDict, Set, Iterable, Optional, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from lifeblood.scheduler import Scheduler


def node_class():
    return ParentChildrenWaiterNode


class ParentChildrenWaiterNode(BaseNode):
    """
    this node will gather tasks from first and second inputs
    when a task from first input has all it's children arriving from the second input
    possibly recursively
    """
    class Entry:
        def __init__(self):
            self.children: Set[int] = set()
            self.parent_ready: bool = False
            self.all_children_dicts: Dict[int, dict] = {}

    @classmethod
    def label(cls) -> str:
        return 'parent-children waiter'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'hierarchy', 'gather', 'wait', 'synchronization', 'barrier', 'child', 'children', 'parent', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'parent_children_waiter'

    def __init__(self, name: str):
        super(ParentChildrenWaiterNode, self).__init__(name)
        self.__cache_children: Dict[int, "ParentChildrenWaiterNode.Entry"] = {}
        self.__main_lock = Lock()
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_input('children')
            ui.add_output('children')
            ui.add_parameter('recursive', 'recursive', NodeParameterType.BOOL, False)
            with ui.multigroup_parameter_block('transfer_attribs'):
                with ui.parameters_on_same_line_block():
                    ui.add_parameter('src_attr_name', 'attribute', NodeParameterType.STRING, 'attr1')
                    ui.add_parameter('transfer_type', 'as', NodeParameterType.STRING, 'extend')\
                        .add_menu((('Extend', 'extend'),
                                   ('Append', 'append'),
                                   ('First', 'first'),
                                   ('Sum', 'sum')))
                    ui.add_parameter('dst_attr_name', 'sort by', NodeParameterType.STRING, 'attr1')
                    ui.add_parameter('sort_by', None, NodeParameterType.STRING, '_builtin_id')
                    ui.add_parameter('reversed', 'reversed', NodeParameterType.BOOL, False)

    def ready_to_process_task(self, task_dict) -> bool:
        context = ProcessingContext(self, task_dict)
        task_id = context.task_field('id')
        children_count = context.task_field('active_children_count')
        if context.task_field('node_input_name') == 'main':
            return children_count == 0 or \
                   task_id in self.__cache_children and children_count == len(self.__cache_children[task_id].children)

        ready: bool = True
        parent_id = context.task_field('parent_id')
        if context.param_value('recursive') and children_count > 0:
            ready = task_id in self.__cache_children and children_count == len(self.__cache_children[task_id].children)
        ready = ready and (
                parent_id not in self.__cache_children or
                task_id not in self.__cache_children[parent_id].children or
                self.__cache_children[parent_id].parent_ready
                )
        return ready

    def process_task(self, context) -> ProcessingResult:
        task_id = context.task_field('id')
        children_count = context.task_field('active_children_count')
        recursive = context.param_value('recursive')

        with self.__main_lock:
            if context.task_field('node_input_name') == 'main':  # parent task
                if children_count == 0:
                    return ProcessingResult()
                if task_id not in self.__cache_children:
                    raise NodeNotReadyToProcess()

                if children_count == len(self.__cache_children[task_id].children):
                    result = ProcessingResult()
                    # transfer attributes
                    num_attribs = context.param_value('transfer_attribs')
                    for i in range(num_attribs):
                        src_attr_name = context.param_value(f'src_attr_name_{i}')
                        transfer_type = context.param_value(f'transfer_type_{i}')
                        dst_attr_name = context.param_value(f'dst_attr_name_{i}')
                        sort_attr_name = context.param_value(f'sort_by_{i}')
                        sort_reversed = context.param_value(f'reversed_{i}')
                        if transfer_type == 'append':
                            gathered_values = []
                            for attribs in sorted(self.__cache_children[task_id].all_children_dicts.values(), key=lambda x: x.get(sort_attr_name, 0), reverse=sort_reversed):
                                if src_attr_name not in attribs:
                                    continue

                                attr_val = attribs[src_attr_name]
                                gathered_values.append(attr_val)
                            result.set_attribute(dst_attr_name, gathered_values)
                        elif transfer_type == 'extend':
                            gathered_values = []
                            for attribs in sorted(self.__cache_children[task_id].all_children_dicts.values(), key=lambda x: x.get(sort_attr_name, 0), reverse=sort_reversed):
                                if src_attr_name not in attribs:
                                    continue

                                attr_val = attribs[src_attr_name]
                                if isinstance(attr_val, list):
                                    gathered_values.extend(attr_val)
                                else:
                                    gathered_values.append(attr_val)
                            result.set_attribute(dst_attr_name, gathered_values)
                        elif transfer_type == 'first':
                            _acd = self.__cache_children[task_id].all_children_dicts
                            if len(_acd) > 0:
                                if sort_reversed:
                                    attribs = max(_acd.values(), key=lambda x: x.get(sort_attr_name, 0))
                                else:
                                    attribs = min(_acd.values(), key=lambda x: x.get(sort_attr_name, 0))
                                if src_attr_name in attribs:
                                    result.set_attribute(dst_attr_name, attribs[src_attr_name])
                        elif transfer_type == 'sum':
                            # we don't care about the order, assume sum is associative
                            gathered_values = None
                            for attribs in self.__cache_children[task_id].all_children_dicts.values():
                                if src_attr_name not in attribs:
                                    continue
                                if gathered_values is None:
                                    gathered_values = attribs[src_attr_name]
                                else:
                                    gathered_values += attribs[src_attr_name]
                            result.set_attribute(dst_attr_name, gathered_values)
                        else:
                            raise NotImplementedError(f'transfer type "{transfer_type}" is not implemented')
                    # release children
                    self.__cache_children[task_id].parent_ready = True
                    return result

                raise NodeNotReadyToProcess()
            else:  # child task
                parent_id = context.task_field('parent_id')
                if parent_id is None:
                    return ProcessingResult(node_output_name='children')
                if recursive and children_count > 0:  # if recursive - we first wait for our children, then add ourselves to our parent children list
                    if task_id not in self.__cache_children:
                        raise NodeNotReadyToProcess()
                    if children_count != len(self.__cache_children[task_id].children):
                        raise NodeNotReadyToProcess()

                if parent_id not in self.__cache_children:
                    self.__cache_children[parent_id] = ParentChildrenWaiterNode.Entry()

                children_attrs_to_pass_up = None
                if self.__cache_children[parent_id].parent_ready:
                    # in recursive case - mark as ready to release children
                    if recursive and task_id in self.__cache_children:  # if we are parent to smth
                        self.__cache_children[task_id].parent_ready = True
                        children_attrs_to_pass_up = self.__cache_children[task_id].all_children_dicts
                    # cleanup
                    self.__cache_children[parent_id].children.remove(task_id)
                    if len(self.__cache_children[parent_id].children) == 0:
                        del self.__cache_children[parent_id]
                    return ProcessingResult(node_output_name='children')

                if task_id not in self.__cache_children[parent_id].children:
                    self.__cache_children[parent_id].children.add(task_id)
                    self.__cache_children[parent_id].all_children_dicts[task_id] = json.loads(context.task_field('attributes'))
                    self.__cache_children[parent_id].all_children_dicts[task_id]['_builtin_id'] = task_id
                    if children_attrs_to_pass_up is not None:
                        self.__cache_children[parent_id].all_children_dicts.update(children_attrs_to_pass_up)
                raise NodeNotReadyToProcess()

        raise NodeNotReadyToProcess()

    def postprocess_task(self, context) -> ProcessingResult:
        res = ProcessingResult()
        res.set_node_output_name(context.task_field('node_output_name', 'main'))
        return res

    def __getstate__(self):
        d = super(ParentChildrenWaiterNode, self).__getstate__()
        assert '_ParentChildrenWaiterNode__main_lock' in d
        del d['_ParentChildrenWaiterNode__main_lock']
        return d



from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment

from typing import Iterable


def node_class():
    return Arnold


class Arnold(BaseNodeWithTaskRequirements):
    @classmethod
    def label(cls) -> str:
        return 'arnold'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'arnold', 'ass', 'stock'

    @classmethod
    def type_name(cls) -> str:
        return 'arnold'

    def __init__(self, name):
        super(Arnold, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.0, 0.54, 0.56)
            ui.add_parameter('ass path', 'ass file path', NodeParameterType.STRING, "`task['file']`")
            ui.add_parameter('image path', 'output image file path', NodeParameterType.STRING, "`task['outimage']`")
            ui.add_parameter('skip if exists', 'skip if result already exists', NodeParameterType.BOOL, False)

            ui.parameter('worker type').set_hidden(True)
            ui.parameter('worker type').set_locked(True)

    def process_task(self, context) -> ProcessingResult:
        args = context.task_attributes()
        if 'file' not in args or 'outimage' not in args:
            raise ProcessingError('required attributes not found')

        env = InvocationEnvironment()
        done_regex = rb'.*?\|\s*(\d+)% done'

        if context.param_value('skip if exists'):
            script = 'import os\n' \
                     'if not os.path.exists({imgpath}):\n' \
                     '    import sys\n' \
                     '    from subprocess import Popen\n' \
                     "    sys.exit(Popen(['kick',  '-dw', '-dp', '-v', '2', '-i', {asspath}, '-o', {imgpath}]).wait())\n" \
                     "else:\n" \
                     "    print('image file already exists, skipping work')\n" \
                    .format(imgpath=repr(context.param_value('image path')),
                            asspath=repr(context.param_value('ass path')))

            invoc = InvocationJob(['python', ':/arnoldcall.py'])
            invoc.set_stdout_progress_regex(done_regex)
            invoc.set_extra_file('arnoldcall.py', script)
        else:
            invoc = InvocationJob(['kick', '-dw', '-dp', '-v', '2', '-i', context.param_value('ass path'), '-o', context.param_value('image path')], env=env)
            invoc.set_stdout_progress_regex(done_regex)
        res = ProcessingResult(invoc)
        return res

    def postprocess_task(self, context) -> ProcessingResult:
        args = context.task_attributes()
        res = ProcessingResult()
        res.set_attribute('file', args['outimage'])
        return res



import json
from types import MappingProxyType
import re

from .config import get_config
from .environment_resolver import EnvironmentResolverArguments

from typing import TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:
    from .basenode import BaseNode
    from .uidata import Parameter


class ProcessingContext:
    class TaskWrapper:
        def __init__(self, task_dict: dict):
            self.__attributes = json.loads(task_dict.get('attributes', '{}'))
            self.__stuff = task_dict

        def __getitem__(self, item):
            return self.__attributes[item]

        def __getattr__(self, item):
            if item in self.__stuff:
                return self.__stuff[item]
            raise AttributeError(f'task has no field {item}')

        def get(self, item, default):
            return self.__attributes.get(item, default)

    class NodeWrapper:
        def __init__(self, node: "BaseNode", context: "ProcessingContext"):
            self.__parameters: Dict[str, "Parameter"] = {x.name(): x for x in node.get_ui().parameters()}
            self.__attrs = {'name': node.name(), 'label': node.label()}
            self.__context = context

        def __getitem__(self, item):
            return self.__parameters[item].value(self.__context)

        def __getattr__(self, item):
            if item in self.__attrs:
                return self.__attrs[item]
            raise AttributeError(f'node has no field {item}')

    class ConfigWrapper:
        def __init__(self, node_type_id):
            self.__config = get_config('scheduler.nodes')
            self.__scheduler_globals = dict(get_config('scheduler').get_option_noasync('scheduler.globals', {}))
            self.__nodetypeid = node_type_id

        def get(self, key, default=None):
            return self.__config.get_option_noasync(f'{self.__nodetypeid}.{key}',
                                                    self.__scheduler_globals.get(key,
                                                                                 default))

        def __getitem__(self, item):
            return self.get(item)

    def __init__(self, node: "BaseNode", task_dict: dict):
        task_dict = dict(task_dict)
        self.__task_attributes = json.loads(task_dict.get('attributes', '{}'))
        self.__task_dict = task_dict
        self.__task_wrapper = ProcessingContext.TaskWrapper(task_dict)
        self.__node_wrapper = ProcessingContext.NodeWrapper(node, self)
        sanitized_name = re.sub(r'\W', lambda m: f'x{ord(m.group(0))}', node.type_name())
        self.__env_args = EnvironmentResolverArguments.deserialize(task_dict.get('environment_resolver_data')) if task_dict.get('environment_resolver_data') is not None else None
        self.__conf_wrapper = ProcessingContext.ConfigWrapper(sanitized_name)
        self.__node = node

    def param_value(self, param_name: str):
        return self.__node.get_ui().parameter(param_name).value(self)

    def locals(self):
        """
        locals to be available during expression evaluation
        node - represents current node
            node['paramname'] returns the value of parameter paramname
            node.name returns node name
            node.label returns node's label
        task - represents task, for which expression is being evaluated
            task['attrname'] returns the value of attrname attribute of current task
            task.fieldname returns task database field called fieldname
        config - general config for this particular node type
            config['entryname'] of config.get('entryname', defaultval) returns entryname from config, or defaultval if entryname does not exist

        :return:
        """
        return {'task': self.__task_wrapper, 'node': self.__node_wrapper, 'config': self.__conf_wrapper}

    def task_attribute(self, attrib_name: str):
        return self.__task_attributes[attrib_name]

    def task_has_attribute(self, attrib_name: str):
        return attrib_name in self.__task_attributes

    def task_attributes(self) -> MappingProxyType:
        return MappingProxyType(self.__task_attributes)

    def task_environment_resolver_arguments(self) -> Optional[EnvironmentResolverArguments]:
        return self.__env_args

    def task_field(self, field_name: str, default_value=None):
        return self.__task_dict.get(field_name, default_value)

    def task_has_field(self, field_name: str):
        return field_name in self.__task_dict

    def task_id(self):
        return self.__task_dict.get('id')

    def task_name(self):
        return self.__task_dict.get('name')



import sys
import os
from lifeblood.basenode import BaseNode, ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.enums import NodeParameterType
from lifeblood.paths import config_path
from typing import Iterable
import subprocess
import time
import json
import tempfile
import shutil


description = \
'''sends any text notification to any matrix room
matrix is a great decentralized communication protocol
you can register accounts on matrix.org or any other publically hosted servers
or self host a server, federated or isolated.

it is recommended to set up token and room once in config
not to expose them in node parameters
to do that - keep default expressions, and 
set values of token and room in your <home>/lifeblood/nodes/config.toml
matrixnotifier.token = 'your_token_here'
matrixnotifier.room = '!your_room_code:server.org'
'''


def node_class():
    return MatrixNotifier


class MatrixNotifier(BaseNode):
    def __init__(self, name):
        super().__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.051, 0.741, 0.545)
            with ui.collapsable_group_block('backend'):
                default_backend = 'fallback'
                if self._is_mc_available():
                    default_backend = 'matrixcommander'
                backend_param = ui.add_parameter('backend', 'backend', NodeParameterType.STRING, default_backend)
                backend_param.add_menu((('fallback', 'fallback'), ('matrix-commander', 'matrixcommander')))
            ui.add_parameter('label_e2ee', None, NodeParameterType.STRING, 'warning: e2e encryption is OFF', can_have_expressions=False, readonly=True)\
                .append_visibility_condition(backend_param, '==', 'fallback')
            ui.add_parameter('token', 'token', NodeParameterType.STRING, '`config["token"]`')\
                .append_visibility_condition(backend_param, '==', 'fallback')
            ui.add_parameter('room', 'room', NodeParameterType.STRING, '`config["room"]`')
            ui.add_parameter('retries', 'retries', NodeParameterType.INT, 5).set_value_limits(value_min=1)
            ui.add_parameter('fail on error', 'fail task on notification sending error', NodeParameterType.BOOL, True)
            with ui.collapsable_group_block('custom server block', 'custom server'):
                ui.add_parameter('server', 'matrix server', NodeParameterType.STRING, '`config.get("server", "https://matrix.org")`')
            ui.add_parameter('message', 'message', NodeParameterType.STRING, '').set_text_multiline()
            with ui.parameters_on_same_line_block():
                ui.add_parameter('do attach', 'attach a file', NodeParameterType.BOOL, False)
                ui.add_parameter('attachment', None, NodeParameterType.STRING, '')

    @classmethod
    def _encode_hs(cls, hs: str):
        """
        given homeserver address return some fs-friendly folder name
        something safe and not giving collisions, but human-readable, not like hash
        """
        return hs.replace('://', '..').replace('/', '_')

    @classmethod
    def _is_mc_available(cls) -> bool:
        mc_cwd = cls.my_plugin().package_data() / 'matrixcommander'
        mc_script = mc_cwd / 'venv' / 'bin' / 'matrix-commander'
        mc_py = mc_cwd / 'venv' / 'bin' / 'python'
        mc_pyexe = mc_cwd / 'venv' / 'bin' / 'python.exe'

        # res = subprocess.Popen([mc_py, mc_script, '--version'], cwd=mc_cwd).wait()
        return mc_cwd.exists() and mc_script.exists() and (mc_py.exists() or mc_pyexe.exists())

    @classmethod
    def label(cls) -> str:
        return 'matrix notifier'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'matrix', 'client', 'notify'

    @classmethod
    def type_name(cls) -> str:
        return 'matrixnotifier'

    @classmethod
    def description(cls) -> str:
        return description

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        print('reporting to matrix')
        retries = context.param_value('retries')
        backend = context.param_value('backend')

        mc_script = None
        mc_py = None
        mc_cwd = None
        mc_conf_base = None
        if backend == 'fallback':
            pass
        elif backend == 'matrixcommander':
            mc_cwd = self.my_plugin().package_data() / 'matrixcommander'
            mc_script = mc_cwd / 'venv' / 'bin' / 'matrix-commander'
            mc_py = mc_cwd / 'venv' / 'bin' / 'python'
            mc_conf_base = config_path(self._encode_hs(context.param_value('server')), 'scheduler.nodes.matrixnotifier.matrixcommander')

            if not mc_py.exists():
                mc_py = mc_cwd / 'venv' / 'bin' / 'python.exe'  # windows case
            if not mc_script.exists():
                mc_script = mc_cwd / 'venv' / 'bin' / 'matrix-commander-script.pyw'  # windows case

            if not mc_py.exists() or not mc_script.exists():
                raise ProcessingError(f'backend "{backend}" was not properly set up. please read the doc')
        else:
            raise ProcessingError(f'unknown backend "{backend}"')

        if retries <= 0:
            retries = sys.maxsize
        for i in range(retries):
            if i > 0:
                time.sleep(2 ** (i - 1))
            print(f'attempt {i+1}...')
            if backend == 'fallback':
                proc = subprocess.Popen([sys.executable, self.my_plugin().package_data() / 'matrixclient.pyz',
                                         'send',
                                         context.param_value('server'),
                                         context.param_value('token'),
                                         context.param_value('room')], stdin=subprocess.PIPE)
                proc.communicate(context.param_value('message').encode('UTF-8'))
            elif backend == 'matrixcommander':
                assert mc_py is not None
                assert mc_script is not None
                assert mc_cwd is not None
                assert mc_conf_base is not None

                proc = subprocess.Popen([mc_py, mc_script,
                                         '--homeserver', context.param_value('server'),
                                         '--room', context.param_value('room'),
                                         '-c', str(mc_conf_base / 'credentials.json'),
                                         '-s', str(mc_conf_base / 'store'),
                                         '-m', '-'],
                                        stdin=subprocess.PIPE,
                                        cwd=mc_cwd,
                                        text=True)
                proc.communicate(context.param_value('message'))
            else:
                raise ProcessingError(f'unknown backend "{backend}"')

            if proc.wait() != 0:
                print(f'attempt {i + 1} failed, sleeping for {2 ** i}')
                continue
            print('reporting to matrix done')
            break
        else:
            print('failed to send notification', file=sys.stderr)
            if context.param_value('fail on error'):
                raise ProcessingError('failed to send notification')

        if context.param_value('do attach'):
            filepath = context.param_value('attachment')
            if not os.path.exists(filepath):
                raise ProcessingError('attachment file does not exist')

            for i in range(retries):
                if i > 0:
                    time.sleep(2 ** (i - 1))
                print(f'attempt {i + 1}...')
                if backend == 'fallback':
                    proc = subprocess.Popen([sys.executable, self.my_plugin().package_data() / 'matrixclient.pyz',
                                             'send',
                                             context.param_value('server'),
                                             context.param_value('token'),
                                             context.param_value('room'),
                                             filepath,
                                             '--message-is-file'])
                elif backend == 'matrixcommander':
                    assert mc_py is not None
                    assert mc_script is not None
                    assert mc_cwd is not None
                    assert mc_conf_base is not None

                    file_flag = '--file'
                    if os.path.splitext(filepath)[1] in ('.jpg', '.jpeg', '.png', '.gif'):
                        file_flag = '--image'

                    proc = subprocess.Popen([mc_py, mc_script,
                                             '--homeserver', context.param_value('server'),
                                             '--room', context.param_value('room'),
                                             '-c', str(mc_conf_base / 'credentials.json'),
                                             '-s', str(mc_conf_base / 'store'),
                                             file_flag, filepath],
                                            stdin=subprocess.DEVNULL,
                                            cwd=mc_cwd)
                    print(proc)
                    print(proc.poll())
                else:
                    raise ProcessingError(f'unknown backend "{backend}"')

                print('about to wait', proc.poll(), mc_cwd)
                if proc.wait() != 0:
                    print(f'attempt {i + 1} failed, sleeping for {2 ** i}')
                    continue
                print('reporting to matrix done')
                break
            else:
                print('failed to send attachment', file=sys.stderr)
                if context.param_value('fail on error'):
                    raise ProcessingError('failed to send attachment')
        return ProcessingResult()




import asyncio
import uuid

from .queue import MessageQueue
from .messages import Message
from .interfaces import MessageReceiverFactory, MessageStreamFactory
from .client import MessageClient, MessageClientFactory, RawMessageClientFactory
from .message_handler import MessageHandlerBase
from .logging import get_logger
from .address import AddressChain, DirectAddress
from .enums import MessageType
from .exceptions import StreamOpeningError
from ..component_base import ComponentBase

from typing import List, Optional, Sequence


class ProcessedSessionsMap:
    def __init__(self):
        self.__map = {}
        self.__empty_event = asyncio.Event()
        self.__empty_event.set()

    def empty_event(self):
        return self.__empty_event

    def __getitem__(self, item):
        return self.__map[item]

    def __setitem__(self, key, value):
        self.__map[key] = value
        if self.__empty_event.is_set():
            self.__empty_event.clear()

    def pop(self, key):
        item = self.__map.pop(key)
        if len(self.__map) == 0:
            self.__empty_event.set()
        return item

    def __contains__(self, item):
        return item in self.__map


class MessageProcessorBase(ComponentBase):
    def __init__(self, listening_address: DirectAddress, *,
                 message_receiver_factory: MessageReceiverFactory,
                 message_stream_factory: MessageStreamFactory,
                 message_client_factory: MessageClientFactory = None,
                 default_client_retry_attempts: Optional[int] = None,
                 message_handlers: Sequence[MessageHandlerBase] = ()):
        super().__init__()
        self.__message_queue = MessageQueue()
        self.__address = listening_address
        self.__sessions_being_processed = ProcessedSessionsMap()
        self.__processing_tasks = set()
        self.__forwarded_messages_count = 0
        self.__message_receiver_factory = message_receiver_factory
        self.__message_stream_factory = message_stream_factory
        self.__message_client_factory = message_client_factory or RawMessageClientFactory()
        self.__default_client_retry_attempts = 2 if default_client_retry_attempts is None else default_client_retry_attempts
        self.__handlers: List[MessageHandlerBase] = list(message_handlers)

        self._logger = get_logger(f'message_processor {type(self).__name__}')

    class _ClientContext:
        def __init__(self, host_address: AddressChain,
                     destination: AddressChain,
                     message_queue: MessageQueue,
                     sessions_being_processed: ProcessedSessionsMap,
                     message_stream_factory: MessageStreamFactory,
                     message_client_factory: MessageClientFactory,
                     force_session: Optional[uuid.UUID] = None,
                     send_retry_attempts: int = 2):
            self.__destination = destination
            self.__force_session = force_session
            self.__sessions_being_processed = sessions_being_processed
            self.__message_queue = message_queue
            self.__message_stream_factory = message_stream_factory
            self.__message_client_factory = message_client_factory
            self.__address = host_address
            self.__session = None
            self.__initialized = False
            self.__send_retry_attempts = send_retry_attempts

        def initialize(self) -> MessageClient:
            if self.__initialized:
                raise RuntimeError('already initialized')
            self.__initialized = True

            if self.__force_session is None:
                while (session := uuid.uuid4()) in self.__sessions_being_processed:
                    pass
            else:
                if self.__force_session in self.__sessions_being_processed:
                    raise ValueError(f'forced session cannot be already in processing! {self.__force_session}')
                session = self.__force_session

            self.__session = session

            client = self.__message_client_factory.create_message_client(
                self.__message_queue, session,
                source_address_chain=self.__address,
                destination_address_chain=self.__destination,
                message_stream_factory=self.__message_stream_factory,
                send_retry_attempts=self.__send_retry_attempts
            )
            self.__sessions_being_processed[session] = client
            return client

        def finalize(self):
            if not self.__initialized:
                raise RuntimeError('not yet initialized')
            self.__sessions_being_processed.pop(self.__session)

        def __enter__(self):
            return self.initialize()

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.finalize()

    def listening_address(self) -> DirectAddress:
        return self.__address

    def forwarded_messages_count(self):
        return self.__forwarded_messages_count

    def add_message_handler(self, handler: MessageHandlerBase):
        if handler not in self.__handlers:
            self.__handlers.append(handler)

    def message_client(self, destination: AddressChain, *, force_session: Optional[uuid.UUID] = None, send_retry_attempts: Optional[int] = None) -> _ClientContext:
        if send_retry_attempts is None:
            send_retry_attempts = self.__default_client_retry_attempts
        return MessageProcessorBase._ClientContext(self.__address,
                                                   destination,
                                                   self.__message_queue,
                                                   self.__sessions_being_processed,
                                                   self.__message_stream_factory,
                                                   self.__message_client_factory,
                                                   force_session,
                                                   send_retry_attempts=send_retry_attempts)

    # @asynccontextmanager
    # async def message_client(self, destination: str, force_session: Optional[uuid.UUID] = None) -> AsyncIterator[MessageClient]:
    #     """
    #     use this line
    #     async with processor.message_client(to_smth) as clinet:
    #         await client.send_message(data)
    #         process_reply(await client.recieve_message())
    #
    #     """
    #     if force_session is None:
    #         while (session := uuid.uuid4()) in self.__sessions_being_processed:
    #             pass
    #     else:
    #         if force_session in self.__sessions_being_processed:
    #             raise ValueError(f'forced session cannot be already in processing! {force_session}')
    #         session = force_session
    #
    #     client = MessageClient(self.__message_queue, session, source=self.__address, destination=destination)
    #     self.__sessions_being_processed[session] = client
    #     try:
    #         yield client
    #     finally:
    #         self.__sessions_being_processed.pop(session)
    #

    def _main_task(self):
        return self.__serve()

    async def __serve(self):
        self._logger.info('starting serving messages')
        server = await self.__message_receiver_factory.create_receiver(self.__address, self.new_message_received)
        self._logger.debug('server started')
        self._main_task_is_ready_now()

        await self._stop_event.wait()

        self._logger.info('message server stopping...')
        self._logger.debug('waiting for all existing sessions to finish')
        await self.__sessions_being_processed.empty_event().wait()
        self._logger.debug('all sessions finished, stopping server')
        await self._pre_receiver_stop()
        server.stop()
        await self._post_receiver_stop()
        await server.wait_till_stopped()
        await self._post_receiver_stop_waited()
        self._logger.info('message server stopped')

    async def _pre_receiver_stop(self):
        return

    async def _post_receiver_stop(self):
        return

    async def _post_receiver_stop_waited(self):
        return

    async def should_process(self, orig_message: Message):
        """
        override this to decide on processing NEW messages
        """
        return not self._stop_event.is_set()

    async def new_message_received(self, message: Message) -> bool:
        r"""
        note about ordering: if session messages come from the same tcp connection - then same protocol instance
         is processing it, so only one message is processed here at a time
         However if messages of same session are coming from multiple tcp connections - there is already no way of
         telling what is the correct order, so avoid that.
         smth like:
         >A1 >A2 <B1 >A3 <B2 <B3 >A4
         \_____/ \_/ \_/ \_____/ \_/
         each group should have single tcp connection, otherwise no guarantee about ordering
        """
        destination = message.message_destination().split_address()
        if destination[0] != self.__address:
            self._logger.error('received message not meant for me, dropping')
            return True

        if len(destination) > 1:  # redirect it further
            dcurrent, dnext = destination[0], destination[1:]
            assert dcurrent == self.__address
            try:
                stream = await self.__message_stream_factory.open_sending_stream(dnext[0], self.__address)
            except StreamOpeningError:
                raise
            except Exception as e:
                raise StreamOpeningError(wrapped_exception=e) from None

            try:
                message.set_message_destination(AddressChain.join_address(dnext))
                message.set_message_source(AddressChain.join_address((dcurrent, *(message.message_source().split_address()))))
                await stream.send_raw_message(message)
                self.__forwarded_messages_count += 1
            finally:
                try:
                    stream.close()
                    await stream.wait_closed()
                except:
                    self._logger.exception('failed to close forwarding stream, suppressing')
            return True

        if message.message_type() == MessageType.SYSTEM_PING:
            return True

        session = message.message_session()
        if session in self.__sessions_being_processed:
            await self.__message_queue.put_message(message)
            return True

        # otherwise - noone is expecting message, so we process it
        # we rely here on that several messages of same session CANNOT be processed here at the same time
        #async with self.message_client(message.message_source(), force_session=session) as client:
        if not await self.should_process(message):
            return False
        context = self.message_client(message.message_source(), force_session=session)
        client = context.initialize()
        task = asyncio.create_task(self.__process_message_wrapper(message, client, context))
        self.__processing_tasks.add(task)
        return True

    async def __process_message_wrapper(self, message: Message, client: MessageClient, context: _ClientContext):
        try:
            for handler in self.__handlers:
                processed = await handler.process_message(message, client)
                if processed:
                    break
            else:
                await self.process_message(message, client)
        except Exception as e:
            self._logger.exception('processing exception happened')
        finally:
            self.__processing_tasks.remove(asyncio.current_task())
            context.finalize()

    async def process_message(self, message: Message, client: MessageClient):
        """
        This will be called only if no handler processed the message

        Override this with actual processing
        """
        self._logger.warning(f'no handlers found to process message "{message}", ignoring')



from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult
from lifeblood.enums import NodeParameterType

from typing import Iterable


def node_class():
    return SetAttributes


class SetAttributes(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'set attributes'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'create', 'set', 'attribute', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'set_attrib'

    def __init__(self, name: str):
        super(SetAttributes, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            with ui.multigroup_parameter_block('attr_count') as wrapper:
                with ui.parameters_on_same_line_block():
                    ui.add_parameter('name', 'Name', NodeParameterType.STRING, '')
                    type_param = ui.add_parameter('type', '', NodeParameterType.INT, 0)
                    type_param.add_menu((('int', NodeParameterType.INT.value),
                                         ('bool', NodeParameterType.BOOL.value),
                                         ('float', NodeParameterType.FLOAT.value),
                                         ('string', NodeParameterType.STRING.value),
                                         ('integer range', -1),
                                         ('dictionary', -2)))

                    ui.add_parameter('svalue', 'val', NodeParameterType.STRING, '').append_visibility_condition(type_param, '==', NodeParameterType.STRING.value)
                    ui.add_parameter('ivalue', 'val', NodeParameterType.INT, 0).append_visibility_condition(type_param, '==', NodeParameterType.INT.value)
                    ui.add_parameter('fvalue', 'val', NodeParameterType.FLOAT, 0.0).append_visibility_condition(type_param, '==', NodeParameterType.FLOAT.value)
                    ui.add_parameter('bvalue', 'val', NodeParameterType.BOOL, False).append_visibility_condition(type_param, '==', NodeParameterType.BOOL.value)
                    # range
                    ui.add_parameter('rf1value', 'to', NodeParameterType.INT, 0).append_visibility_condition(type_param, '==', -1)
                    ui.add_parameter('rf2value', 'step', NodeParameterType.INT, 9).append_visibility_condition(type_param, '==', -1)
                    ui.add_parameter('rf3value', '', NodeParameterType.INT, 1).append_visibility_condition(type_param, '==', -1)
                #
                # dictionary (no recursive yet)
                with ui.multigroup_parameter_block('dict_attrs'):
                    with ui.parameters_on_same_line_block():
                        ui.add_parameter('dict_item_name', 'Name', NodeParameterType.STRING, '')
                        dict_type_param = ui.add_parameter('dict_item_type', '', NodeParameterType.INT, 0)
                        dict_type_param.add_menu((('int', NodeParameterType.INT.value),
                                                  ('bool', NodeParameterType.BOOL.value),
                                                  ('float', NodeParameterType.FLOAT.value),
                                                  ('string', NodeParameterType.STRING.value),))

                        ui.add_parameter('dict_item_svalue', 'val', NodeParameterType.STRING, '').append_visibility_condition(dict_type_param, '==', NodeParameterType.STRING.value)
                        ui.add_parameter('dict_item_ivalue', 'val', NodeParameterType.INT, 0).append_visibility_condition(dict_type_param, '==', NodeParameterType.INT.value)
                        ui.add_parameter('dict_item_fvalue', 'val', NodeParameterType.FLOAT, 0.0).append_visibility_condition(dict_type_param, '==', NodeParameterType.FLOAT.value)
                        ui.add_parameter('dict_item_bvalue', 'val', NodeParameterType.BOOL, False).append_visibility_condition(dict_type_param, '==', NodeParameterType.BOOL.value)
                #
                wrapper.multigroup().parameter('dict_attrs').append_visibility_condition(type_param, '==', -2)

            ui.color_scheme().set_main_color(0.15, 0.24, 0.25)
        # # Example how one would initialize multiblock to have initial nonzero value
        # multiblock = list(ui.items())[-1]
        # assert isinstance(multiblock, MultiGroupLayout)  # TODO: find a better workflow of getting that last layout added
        # multiblock.add_template_instance()
        # multiblock.add_template_instance()

        # def _printone(level, i=0):
        #     if isinstance(level, Parameter):
        #         print('\t'*(i-1), f'====parameter {level.name()}')
        #         return
        #     for item in level.items(recursive=False):
        #         print('\t'*i, item)
        #         _printone(item, i+1)

        #_printone(ui.main_parameter_layout())

    def process_task(self, context) -> ProcessingResult:
        attr_count = context.param_value('attr_count')
        res = ProcessingResult()
        for i in range(attr_count):
            attr_name = context.param_value(f'name_{i}')
            if attr_name == '':  # however some spaces - is a valid parameter name
                continue
            attr_type = context.param_value(f'type_{i}')
            if attr_type == NodeParameterType.INT.value:
                attr_val = context.param_value(f'ivalue_{i}')
            elif attr_type == NodeParameterType.BOOL.value:
                attr_val = context.param_value(f'bvalue_{i}')
            elif attr_type == NodeParameterType.FLOAT.value:
                attr_val = context.param_value(f'fvalue_{i}')
            elif attr_type == NodeParameterType.STRING.value:
                attr_val = context.param_value(f'svalue_{i}')
            elif attr_type == -1:
                attr_val = list(range(context.param_value(f'rf1value_{i}'), context.param_value(f'rf2value_{i}') + 1, context.param_value(f'rf3value_{i}')))
            else:
                raise NotImplementedError('unexpected type')
            res.set_attribute(attr_name, attr_val)
        return res

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



# this is a helper module that is SUPPOSED to be separate from lifeblood
# this module should be used by DCCs and other scripts that want to communicate with lifeblood
# but this module does NOT use lifeblood module directly, even if that means code duplication
# the point is to have some small light module without any dependencies
# that can be imported to a variety of DCCs and not cause any conflicts etc etc



import asyncio
import uuid
from .messages import Message
from .queue import MessageQueue
from .address import AddressChain
from .interfaces import MessageStreamFactory
from .exceptions import MessageSendingError, MessageTransferTimeoutError, MessageReceiveTimeoutError
from .logging import get_logger

from typing import Optional


class MessageClient:
    logger = get_logger('message_client')

    def __init__(self, queue: MessageQueue, session: uuid.UUID, *,
                 source_address_chain: AddressChain,
                 destination_address_chain: AddressChain,
                 message_stream_factory: MessageStreamFactory,
                 send_retry_attempts: int = 2):
        self.__message_queue = queue
        self.__last_sent_message: Optional[Message] = None
        self.__session: uuid.UUID = session
        self.__message_stream_factory = message_stream_factory

        self.__source = source_address_chain.split_address()
        self.__source_str = source_address_chain

        self.__destination = destination_address_chain.split_address()
        self.__destination_str = destination_address_chain

        self.__attempts = send_retry_attempts
        self.__init_timeout = 1

    def set_base_attempt_timeout(self, timeout: float):
        if timeout <= 0:
            self.logger.warning('cannot set base attempt timeout to value <= 0, reset to default')
            timeout = 1
        self.__init_timeout = timeout

    def session(self) -> uuid.UUID:
        return self.__session

    async def send_message(self, data: bytes) -> Message:
        last_exception = None
        timeout = self.__init_timeout

        for attempt in range(self.__attempts):
            stream = None
            try:
                try:
                    stream = await self.__message_stream_factory.open_sending_stream(self.__destination[0], self.__source[0])
                except OSError as e:
                    raise MessageSendingError(wrapped_exception=e) from None
                message = await stream.send_data_message(data, self.__destination_str, source=self.__source_str, session=self.__session)
                self.__last_sent_message = message
                return message
            except MessageTransferTimeoutError as e:
                self.logger.warning('timeout while trying to send message')
                last_exception = e
                # TODO: should we allow more attempts in this case? kinda like unexpected multiplier for the timeout
            except MessageSendingError as e:
                self.logger.warning(f'failed to send message: {str(e.wrapped_exception())}')
                last_exception = e
            finally:
                if stream is None:
                    self.logger.warning(f'failed to initialize stream: {str(last_exception)}')
                else:
                    stream.close()
                    try:
                        await stream.wait_closed()
                    except Exception as e:
                        self.logger.warning(f'was unable to properly close stream: {str(e)}')
            if attempt < self.__attempts-1:
                self.logger.warning(f'unable to sent message, will retry in {timeout}')
                await asyncio.sleep(timeout)
                timeout *= 2
        else:
            self.logger.error(f'unable to sent message, out of attempts, failing.')
            raise last_exception

    async def receive_message(self, timeout: Optional[float] = 90) -> Message:
        task = self.__message_queue.get_message(self.__session)
        if timeout is None:
            message = await task
        else:
            try:
                message = await asyncio.wait_for(task, timeout=timeout)
            except asyncio.TimeoutError as e:
                raise MessageReceiveTimeoutError(wrapped_exception=e) from None
        return message


class MessageClientFactory:
    def create_message_client(self, queue: MessageQueue, session: uuid.UUID, *,
                              source_address_chain: AddressChain,
                              destination_address_chain: AddressChain,
                              message_stream_factory: MessageStreamFactory,
                              send_retry_attempts: int = 2) -> MessageClient:
        raise NotImplementedError()


class RawMessageClientFactory(MessageClientFactory):
    def create_message_client(self, queue: MessageQueue, session: uuid.UUID, *,
                              source_address_chain: AddressChain,
                              destination_address_chain: AddressChain,
                              message_stream_factory: MessageStreamFactory,
                              send_retry_attempts: int = 2) -> MessageClient:
        return MessageClient(queue, session,
                             source_address_chain=source_address_chain,
                             destination_address_chain=destination_address_chain,
                             message_stream_factory=message_stream_factory,
                             send_retry_attempts=send_retry_attempts)



import socket
import struct
import pickle
import json
import time
from .nethelpers import recv_exactly, send_string, recv_string
from .environment_resolver import EnvironmentResolverArguments

try:
    from typing import Optional, Tuple, List
except ImportError:  # py2 does not have typing module, but adequete IDE will still pick it up
    pass


class scheduler_client:
    def __init__(self, ip, port):
        self.__addr = (ip, port)
        self.__socket = None  # type: Optional[socket.socket]

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def connect(self):
        if self.__socket is not None:
            return

        addr, sport = self.__addr
        port = int(sport)
        self.__socket = socket.create_connection((addr, port), timeout=30)
        self.__socket.sendall(b'\0\0\0\0')

    def close(self):
        if self.__socket is None:
            return
        self.__socket.close()

    def find_nodes_by_name(self, node_name):  # type: (str) -> Tuple[int, ...]
        send_string(self.__socket, 'nodenametoid')

        send_string(self.__socket, node_name)
        num = struct.unpack('>Q', recv_exactly(self.__socket, 8))[0]
        ids = struct.unpack('>' + 'Q'*num, recv_exactly(self.__socket, 8*num))
        return ids

    def find_tasks_by_name(self, task_name):  # type: (str) -> Tuple[int, ...]
        send_string(self.__socket, 'tasknametoid')

        send_string(self.__socket, task_name)
        num = struct.unpack('>Q', recv_exactly(self.__socket, 8))[0]
        ids = struct.unpack('>' + 'Q'*num, recv_exactly(self.__socket, 8*num))
        return ids

    def get_task_attributes(self, task_id):
        attribs, _ = self.get_task_attributes_and_env(task_id)
        return attribs

    def get_task_environment_resolver_arguments(self, task_id):
        attribs, _ = self.get_task_attributes_and_env(task_id)
        return attribs

    def get_task_attributes_and_env(self, task_id):
        send_string(self.__socket, 'gettaskattribs')

        self.__socket.sendall(struct.pack('>Q', task_id))
        data_len = struct.unpack('>Q', recv_exactly(self.__socket, 8))[0]
        data_attribs = recv_exactly(self.__socket, data_len)
        data_len = struct.unpack('>Q', recv_exactly(self.__socket, 8))[0]
        data_env = recv_exactly(self.__socket, data_len)

        attribs = pickle.loads(data_attribs)
        env = EnvironmentResolverArguments.deserialize(data_env)

        return attribs, env

    def get_task_fields(self, task_id):
        send_string(self.__socket, 'gettaskstate')

        self.__socket.sendall(struct.pack('>Q', task_id))
        data_len = struct.unpack('>Q', recv_exactly(self.__socket, 8))[0]
        data = recv_exactly(self.__socket, data_len)

        return json.loads(data)


class Task:
    class TaskState:  # note, this is NOT a ENUM, to keep py2 compatibility without extra deps
        WAITING = 0  # arrived at node, does not know what to do
        GENERATING = 1  # node is generating work load
        READY = 2  # ready to be scheduled
        INVOKING = 11  # task is being switched to IN_PROGRESS
        IN_PROGRESS = 3  # is being worked on by a worker
        POST_WAITING = 4  # task is waiting to be post processed by node
        POST_GENERATING = 5  # task is being post processed by node
        DONE = 6  # done, needs further processing
        ERROR = 7  # some internal error, not allowing to process task. NOT INVOCATION ERROR
        SPAWNED = 8  # spawned tasks are just passed down from node's "spawned" output
        DEAD = 9  # task will not be processed any more
        SPLITTED = 10  # task has been splitted, and will remain idle until splits are gathered

    def __init__(self, server_address, task_id):  # type: (Tuple[str, int], int) -> None
        self.__id = task_id
        self.__name = ''
        self.__parent_id = -1
        self.__children_count = -1
        self.__active_children_count = -1
        self.__split_level = -1
        self.__dead = False
        self.__paused = False
        self.__state = Task.TaskState.ERROR

        self.__fields_update_timestamp = 0
        self.__attribs_update_timestamp = 0

        self.__attribs = {}
        self.__env_resolver_args = None

        self.__cache_validity_timeout = 2
        self.__addr = server_address

    def _ensure_fields_fetched(self):
        if time.time() - self.__fields_update_timestamp < self.__cache_validity_timeout:
            return
        with scheduler_client(*self.__addr) as client:
            data_dict = client.get_task_fields(self.__id)
        self.__id = data_dict['id']
        self.__name = data_dict['name']
        self.__parent_id = data_dict['parent_id']
        self.__children_count = data_dict['children_count']
        self.__active_children_count = data_dict['active_children_count']
        self.__split_level = data_dict['split_level']
        self.__dead = bool(data_dict['dead'])
        self.__paused = bool(data_dict['paused'])
        self.__state = data_dict['state']
        self.__fields_update_timestamp = time.time()

    def _ensure_attributes_fetched(self):
        if time.time() - self.__attribs_update_timestamp < self.__cache_validity_timeout:
            return
        with scheduler_client(*self.__addr) as client:
            self.__attribs, self.__env_resolver_args = client.get_task_attributes_and_env(self.__id)
        self.__attribs_update_timestamp = time.time()

    @property
    def id(self):
        self._ensure_fields_fetched()
        return self.__id

    @property
    def name(self):
        self._ensure_fields_fetched()
        return self.__name

    @property
    def parent_id(self):
        self._ensure_fields_fetched()
        return self.__parent_id

    @property
    def children_count(self):
        self._ensure_fields_fetched()
        return self.__children_count

    @property
    def active_children_count(self):
        self._ensure_fields_fetched()
        return self.__active_children_count

    @property
    def split_level(self):
        self._ensure_fields_fetched()
        return self.__split_level

    @property
    def dead(self):
        self._ensure_fields_fetched()
        return self.__dead

    @property
    def paused(self):
        self._ensure_fields_fetched()
        return self.__paused

    @property
    def state(self):
        self._ensure_fields_fetched()
        return self.__state

    def get_attributes(self):
        self._ensure_attributes_fetched()
        return self.__attribs

    def get_attribute_names(self):
        self._ensure_attributes_fetched()
        return tuple(self.__attribs.keys())

    def get_attribute_value(self, attrib_name):
        self._ensure_attributes_fetched()
        return self.__attribs[attrib_name]

    def get_environment_resolver_arguments(self):
        self._ensure_attributes_fetched()
        return self.__env_resolver_args


def get_task_by_id(scheduler_address, task_id):  # type: (Tuple[str, int], int) -> Task
    return Task(scheduler_address, task_id)


def get_tasks_by_name(scheduler_address, task_name):  # type: (Tuple[str, int], str) -> List[Task]
    with scheduler_client(*scheduler_address) as client:
        return [Task(scheduler_address, x) for x in client.find_tasks_by_name(task_name)]



import struct
import pickle
import json
import asyncio
from asyncio.exceptions import IncompleteReadError
from . import logging
from .uidata import NodeUi, Parameter, ParameterLocked, ParameterReadonly, ParameterNotFound, ParameterCannotHaveExpressions
from .ui_protocol_data import NodeGraphStructureData, TaskGroupBatchData, TaskBatchData, WorkerBatchData, UiData, InvocationLogData, IncompleteInvocationLogData
from .ui_events import TaskEvent
from .enums import NodeParameterType, TaskState, SpawnStatus, TaskGroupArchivedState
from .exceptions import NotSubscribedError
from . import pluginloader
from .invocationjob import InvocationJob
from .net_classes import NodeTypeMetadata
from .taskspawn import NewTask
from .snippets import NodeSnippetData, NodeSnippetDataPlaceholder
from .environment_resolver import EnvironmentResolverArguments
from .buffered_connection import BufferedConnection

from typing import Any, Dict, Iterable, TYPE_CHECKING, Optional, Tuple, List, Union
if TYPE_CHECKING:
    from .basenode import BaseNode
    from .scheduler import Scheduler


def _serialize_json_dict(d: dict) -> bytes:
    return json.dumps(d).encode('UTF-8')


def _deserialize_json_dict(data: bytes) -> dict:
    return json.loads(data.decode('UTF-8'))


class SchedulerUiProtocol(asyncio.StreamReaderProtocol):
    def __init__(self, scheduler):
        self.__logger = logging.get_logger('scheduler.uiprotocol')
        self.__scheduler: "Scheduler" = scheduler
        self.__reader = asyncio.StreamReader()
        self.__timeout = 60.0
        self.__saved_references = []
        super(SchedulerUiProtocol, self).__init__(self.__reader, self.connection_cb)

    async def connection_cb(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        # there is a bug in py <=3.8, callback task can be GCd
        # see https://bugs.python.org/issue46309
        # so we HAVE to save a reference to self somewhere
        self.__saved_references.append(asyncio.current_task())

        #
        #
        # commands
        # async def comm_get_full_state():  # if command == b'getfullstate':
        #     task_groups = []
        #     skip_dead, skip_archived_groups = struct.unpack('>??', await reader.readexactly(2))
        #     for i in range(struct.unpack('>I', await reader.readexactly(4))[0]):
        #         task_groups.append(await read_string())
        #
        #     uidata = await self.__scheduler.ui_state_access.get_full_ui_state(task_groups, skip_dead=skip_dead, skip_archived_groups=skip_archived_groups)
        #     await uidata.serialize_to_streamwriter(writer)

        async def comm_get_db_uid():
            writer.write(struct.pack('>Q', self.__scheduler.db_uid()))

        async def comm_get_ui_graph_state_update_id():  # get_ui_graph_state_update_id
            writer.write(struct.pack('>Q', self.__scheduler.ui_state_access.graph_update_id))

        async def comm_get_ui_graph_state():  # get_ui_graph_state
            state = await self.__scheduler.ui_state_access.get_nodes_ui_state()
            update_id = self.__scheduler.ui_state_access.graph_update_id  # there should be no possibility for race condition between aquiring of state and update_id, cus async, but not multithreaded
            writer.write(struct.pack('>Q', update_id))
            await state.serialize_to_streamwriter(writer)

        async def comm_get_ui_task_groups():  # get_ui_task_groups
            skip_archived_groups, = struct.unpack('>?', await reader.readexactly(1))
            state = await self.__scheduler.ui_state_access.get_task_groups_ui_state(fetch_statistics=True, skip_archived_groups=skip_archived_groups)
            await state.serialize_to_streamwriter(writer)

        async def comm_get_ui_tasks_state():  # get_ui_tasks_state
            include_dead, num_groups = struct.unpack('>?Q', await reader.readexactly(9))
            groups = []
            for _ in range(num_groups):
                groups.append(await read_string())
            state = await self.__scheduler.ui_state_access.get_tasks_ui_state(groups, not include_dead)
            await state.serialize_to_streamwriter(writer)

        # ui events

        async def comm_request_subscribe_to_task_events():  # request_task_events
            include_dead, num_groups, logging_time = struct.unpack('>?Qd', await reader.readexactly(17))
            groups = []
            for _ in range(num_groups):
                groups.append(await read_string())
            events = await self.__scheduler.ui_state_access.subscribe_to_task_events_for_groups(groups, not include_dead, logging_time)
            writer.write(struct.pack('>Q', len(events)))
            for event in events:
                await event.serialize_to_streamwriter(writer)

        async def comm_request_task_events_since_id():  # task_events_since_id
            include_dead, num_groups, last_known_event_id = struct.unpack('>?QQ', await reader.readexactly(17))
            groups = []
            for _ in range(num_groups):
                groups.append(await read_string())
            try:
                events = await self.__scheduler.ui_state_access.get_events_for_groups_since_event_id(groups, not include_dead, last_known_event_id)
            except NotSubscribedError:
                writer.write(struct.pack('>?', False))
                return

            writer.write(struct.pack('>?Q', True, len(events)))
            for event in events:
                await event.serialize_to_streamwriter(writer)

        # end of ui events

        async def comm_get_ui_workers_state():  # get_ui_workers_state
            state = await self.__scheduler.ui_state_access.get_workers_ui_state()
            await state.serialize_to_streamwriter(writer)

        async def comm_get_invoc_meta():  # elif command in (b'getinvocmeta', b'getlog', b'getalllog'):
            # if command == b'getinvocmeta':
            task_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            all_meta = await self.__scheduler.get_invocation_metadata(task_id)
            writer.write(struct.pack('>Q', len(all_meta)))
            for node_id, invoc_list in all_meta.items():
                writer.write(struct.pack('>QQ', node_id, len(invoc_list)))
                for log in invoc_list:
                    await log.serialize_to_streamwriter(writer)

        async def comm_get_log():
            # elif command == b'getlog':
            invocation_id, = struct.unpack('>Q', await reader.readexactly(8))
            log = await self.__scheduler.get_log(invocation_id)
            writer.write(struct.pack('>?', log is not None))
            if log is not None:
                await log.serialize_to_streamwriter(writer)

        # brings in interface data for one particular node
        async def comm_get_node_interface():  # elif command == b'getnodeinterface':
            node_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            async with self.__scheduler.node_object_by_id_for_reading(node_id) as node:
                nodeui: NodeUi = node.get_ui()
                data: bytes = await nodeui.serialize_async()
            writer.write(struct.pack('>I', len(data)))
            writer.write(data)

        async def comm_get_task_attribs():  # elif command == b'gettaskattribs':
            task_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            attribs, env_attribs = await self.__scheduler.get_task_attributes(task_id)

            data_attirbs: bytes = await asyncio.get_event_loop().run_in_executor(None, _serialize_json_dict, attribs)
            data_env: bytes = b''
            if env_attribs is not None:
                data_env: bytes = await EnvironmentResolverArguments.serialize_async(env_attribs)
            writer.write(struct.pack('>Q', len(data_attirbs)))
            writer.write(data_attirbs)
            writer.write(struct.pack('>Q', len(data_env)))
            writer.write(data_env)

        async def comm_get_task_invocation():  # elif command == b'gettaskinvoc':
            task_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            data = await self.__scheduler.get_task_invocation_serialized(task_id)
            if data is None:
                writer.write(struct.pack('>Q', 0))
            else:
                writer.write(struct.pack('>Q', len(data)))
                writer.write(data)

        # node related commands
        async def comm_list_node_types():  # elif command == b'listnodetypes':
            typemetas = []
            for type_name, module in pluginloader.plugins.items():
                cls = module.node_class()
                typemetas.append(NodeTypeMetadata(cls))
            writer.write(struct.pack('>Q', len(typemetas)))
            for typemeta in typemetas:
                data: bytes = await asyncio.get_event_loop().run_in_executor(None, pickle.dumps, typemeta)
                writer.write(struct.pack('>Q', len(data)))
                writer.write(data)

        async def comm_list_presets():
            preset_metadata: Dict[str, Dict[str, NodeSnippetDataPlaceholder]] = {pack: {pres: NodeSnippetDataPlaceholder.from_nodesnippetdata(snip) for pres, snip in packdata.items()} for pack, packdata in pluginloader.presets.items()}
            data: bytes = await asyncio.get_event_loop().run_in_executor(None, pickle.dumps, preset_metadata)
            writer.write(struct.pack('>Q', len(data)))
            writer.write(data)

        async def comm_get_node_preset():
            package_name = await read_string()
            preset_name = await read_string()
            if package_name not in pluginloader.presets or preset_name not in pluginloader.presets[package_name]:
                self.__logger.warning(f'requested preset {package_name}::{preset_name} is not found')
                writer.write(struct.pack('>?', False))
                return
            writer.write(struct.pack('>?', True))
            data = pluginloader.presets[package_name][preset_name].serialize(ascii=False)
            writer.write(struct.pack('>Q', len(data)))
            writer.write(data)

        async def comm_remove_node():  # elif command == b'removenode':
            node_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            deleted = await self.__scheduler.remove_node(node_id)
            writer.write(b'\1' if deleted else b'\0')

        async def comm_add_node():  # elif command == b'addnode':
            node_type = await read_string()
            node_name = await read_string()
            node_id = await self.__scheduler.add_node(node_type, node_name)
            writer.write(struct.pack('>Q', node_id))

        async def comm_wipe_node():  # elif command == b'wipenode':
            node_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            await self.__scheduler.wipe_node_state(node_id)
            writer.write(b'\1')

        async def comm_node_has_param():  # elif command == b'nodehasparam':
            node_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            param_name = await read_string()
            try:
                async with self.__scheduler.node_object_by_id_for_reading(node_id) as node:
                    if node.param(param_name) is None:
                        writer.write(b'\0')
                    else:
                        writer.write(b'\1')
            except Exception:
                self.__logger.warning(f'FAILED get node {node_id}')
                writer.write(b'\0')

        async def comm_set_node_param():  # elif command == b'setnodeparam':
            node_id, param_type, has_expression = struct.unpack('>QI?', await reader.readexactly(13))
            param_name = await read_string()
            param_expr = None
            if has_expression:
                param_expr = await read_string()
            if param_type == NodeParameterType.FLOAT.value:
                param_value = struct.unpack('>d', await reader.readexactly(8))[0]
            elif param_type == NodeParameterType.INT.value:
                param_value = struct.unpack('>q', await reader.readexactly(8))[0]
            elif param_type == NodeParameterType.BOOL.value:
                param_value = struct.unpack('>?', await reader.readexactly(1))[0]
            elif param_type == NodeParameterType.STRING.value:
                param_value = await read_string()
            else:
                raise NotImplementedError()
            try:
                async with self.__scheduler.node_object_by_id_for_writing(node_id) as node:  # type: BaseNode
                    param = await asyncio.get_event_loop().run_in_executor(None, node.param, param_name)
                    # first set expression
                    if param.has_expression() and not has_expression:
                        await asyncio.get_event_loop().run_in_executor(None, param.remove_expression)
                    elif has_expression:
                        assert param_expr is not None
                        await asyncio.get_event_loop().run_in_executor(None, param.set_expression, param_expr)
                    # TODO: if error happens below - we'll end up with half set parameter... not nice
                    # then set value (value can be set independent, remember)
                    await asyncio.get_event_loop().run_in_executor(None, param.set_value, param_value)
                    value = param.unexpanded_value()
            except ParameterReadonly:
                self.__logger.warning(f'failed request to set node {node_id} parameter "{param_name}"({NodeParameterType(param_type).name}). parameter is READ ONLY')
                writer.write(b'\0')
            except ParameterLocked:
                self.__logger.warning(f'failed request to set node {node_id} parameter "{param_name}"({NodeParameterType(param_type).name}). parameter is LOCKED')
                writer.write(b'\0')
            except Exception as e:
                err_val_prev = str(param_value)
                if len(err_val_prev) > 23:
                    err_val_prev = err_val_prev[:20] + '...'
                self.__logger.warning(f'FAILED request to set node {node_id} parameter "{param_name}"({NodeParameterType(param_type).name}) to {err_val_prev}')
                self.__logger.exception(e)
                writer.write(b'\0')
            else:
                writer.write(b'\1')
                #
                # send back actual result
                if param_type == NodeParameterType.FLOAT.value:
                    writer.write(struct.pack('>d', value))
                elif param_type == NodeParameterType.INT.value:
                    writer.write(struct.pack('>q', value))
                elif param_type == NodeParameterType.BOOL.value:
                    writer.write(struct.pack('>?', value))
                elif param_type == NodeParameterType.STRING.value:
                    await write_string(value)
                else:
                    raise NotImplementedError()

        async def comm_set_node_params():  # elif command == b'batchsetnodeparams':
            node_id, param_count, want_result = struct.unpack('>QQ?', await reader.readexactly(17))

            stuff_to_set: List[Tuple[str, NodeParameterType, bool, Union[float, int, bool, str]]] = []
            return_order: List[str] = []
            for _ in range(param_count):
                param_name = await read_string()
                param_type_raw, value_is_expression = struct.unpack('>I?', await reader.readexactly(5))
                param_type_raw: int
                value_is_expression: bool

                if value_is_expression:
                    param_value = await read_string()
                elif param_type_raw == NodeParameterType.FLOAT.value:
                    param_value = struct.unpack('>d', await reader.readexactly(8))[0]
                elif param_type_raw == NodeParameterType.INT.value:
                    param_value = struct.unpack('>q', await reader.readexactly(8))[0]
                elif param_type_raw == NodeParameterType.BOOL.value:
                    param_value = struct.unpack('>?', await reader.readexactly(1))[0]
                elif param_type_raw == NodeParameterType.STRING.value:
                    param_value = await read_string()
                else:
                    raise NotImplementedError()
                stuff_to_set.append((param_name, NodeParameterType(param_type_raw), value_is_expression, param_value))
                return_order.append(param_name)

            # now it is possible that some parameters won't exist before other parameters are set
            #  for example: elements of multigroup won't be available until multigroup's count is set
            #  also multigroup params may be nested, so it may be complicated to untangle beforehand
            #  but there may be other
            return_values: Dict[str, Tuple[bool, Optional[Tuple[NodeParameterType, Union[float, int, bool, str]]]]] = {}  # list of param_names to [set success?, final value]
            async with self.__scheduler.node_object_by_id_for_writing(node_id) as node:
                node: BaseNode
                cycle_start = None
                something_was_set_last_loop = False
                for param_name, param_type, value_is_expression, param_value in stuff_to_set:
                    try:
                        await asyncio.get_event_loop().run_in_executor(None, node.param, param_name)
                    except ParameterNotFound:
                        # some parameters may only be present after other parameters are set.
                        #  for example - multigroup element will appear only after count is set
                        #  so we postpone setting missing parameters
                        #  and only consider it a failure if no more parameters can be found at all
                        if cycle_start is None or cycle_start == param_name and something_was_set_last_loop:
                            cycle_start = param_name
                            something_was_set_last_loop = False
                            stuff_to_set.append((param_name, param_type, value_is_expression, param_value))  # put back
                        else:  # means we already checked all parameters since last successful set and set nothing new
                            self.__logger.warning(f'failed request to set node {node_id} parameter "{param_name}" - parameter not found')
                            return_values[param_name] = (False, None)
                        continue
                    try:
                        def _set_helper(node: "BaseNode", param_name: str, value_is_expression: bool, value: Union[float, int, bool, str]):
                            param = node.param(param_name)
                            if value_is_expression:
                                param.set_expression(param_value)
                            else:
                                if param.has_expression():
                                    param.set_expression(None)
                                param.set_value(value)

                        await asyncio.get_event_loop().run_in_executor(None, _set_helper, node, param_name, value_is_expression, param_value)
                        something_was_set_last_loop = True
                        value = node.param(param_name).unexpanded_value()
                    except ParameterReadonly:
                        self.__logger.warning(f'failed request to set node {node_id} parameter "{param_name}"({NodeParameterType(param_type).name}). parameter is READ ONLY')
                        return_values[param_name] = (False, None)
                    except ParameterLocked:
                        self.__logger.warning(f'failed request to set node {node_id} parameter "{param_name}"({NodeParameterType(param_type).name}). parameter is LOCKED')
                        return_values[param_name] = (False, None)
                    except ParameterCannotHaveExpressions:
                        self.__logger.warning(f'failed request to set node {node_id} parameter "{param_name}" expression. parameter cannot have expressions')
                        return_values[param_name] = (False, None)
                    except Exception as e:
                        err_val_prev = str(param_value)
                        if len(err_val_prev) > 23:
                            err_val_prev = err_val_prev[:20] + '...'
                        self.__logger.warning(f'FAILED request to set node {node_id} parameter "{param_name}"({NodeParameterType(param_type).name}) to {err_val_prev}')
                        self.__logger.exception(e)
                        return_values[param_name] = (False, None)
                    else:  # all went well, parameter was set
                        return_values[param_name] = (True, (param_type, value))

            if not want_result:
                writer.write(b'\1')
            else:
                for param_name in return_order:
                    was_set, other = return_values[param_name]
                    if not was_set:
                        writer.write(b'\0')
                        continue
                    writer.write(b'\1')
                    param_type, value = other
                    if param_type == NodeParameterType.FLOAT:
                        writer.write(struct.pack('>d', value))
                    elif param_type == NodeParameterType.INT:
                        writer.write(struct.pack('>q', value))
                    elif param_type == NodeParameterType.BOOL:
                        writer.write(struct.pack('>?', value))
                    elif param_type == NodeParameterType.STRING:
                        await write_string(value)
                    else:  # only possible if we seriously forgot to update this protocol after updating NodeParameterType or smth
                        raise NotImplementedError()
            # note: values written back to sender is:  \? (value) \? (value) ... \? (value)
            #       where \? is \1 if value was set, and \0 if wasn't

        # async def comm_set_node_param_expression():  # elif command == b'setnodeparamexpression':
        #     node_id, set_or_unset = struct.unpack('>Q?', await reader.readexactly(9))
        #     param_name = await read_string()
        #     res = b'\1'
        #     try:
        #         async with self.__scheduler.node_object_by_id_for_writing(node_id) as node:
        #             if set_or_unset:  # means SET
        #                 expression = await read_string()
        #                 await asyncio.get_event_loop().run_in_executor(None, node.param(param_name).set_expression, expression)
        #                 # node.param(param_name).set_expression(expression)
        #             else:
        #                 await asyncio.get_event_loop().run_in_executor(None, node.param(param_name).remove_expression)
        #     except Exception as e:
        #         self.__logger.error(f'error setting parameter expression on {node_id}: {str(e)}')
        #         res = b'\0'
        #     writer.write(res)

        async def comm_apply_node_settings():
            node_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            settings_name = await read_string()
            try:
                async with self.__scheduler.node_object_by_id_for_writing(node_id) as node:
                    await asyncio.get_event_loop().run_in_executor(None, node.apply_settings, settings_name)
            except Exception:
                self.__logger.exception(f'FAILED to apply node settings for node {node_id}, settings name "{settings_name}"')
                writer.write(b'\0')
            else:
                writer.write(b'\1')

        async def comm_save_custom_node_settings():
            node_type_name = await read_string()
            settings_name = await read_string()
            datasize = struct.unpack('>Q', await reader.readexactly(8))[0]

            settings = await asyncio.get_event_loop().run_in_executor(None, pickle.loads, await reader.readexactly(datasize))
            try:
                pluginloader.add_settings_to_existing_package('custom_default', node_type_name, settings_name, settings)
            except RuntimeError as e:
                self.__logger.error(f'failed to add custom node settings: {str(e)}')
                writer.write(b'\0')
            else:
                writer.write(b'\1')

        async def comm_set_settings_default():
            node_type_name = await read_string()
            is_setting = struct.unpack('>?', await reader.readexactly(1))[0]
            if is_setting:
                settings_name = await read_string()
            else:
                settings_name = None
            try:
                pluginloader.set_settings_as_default(node_type_name, settings_name)
            except RuntimeError:
                self.__logger.error(f'failed to set node default settings: {str(e)}')
                writer.write(b'\0')
            else:
                writer.write(b'\1')

        async def comm_rename_node():  # if command == b'renamenode':
            node_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            node_name = await read_string()
            name = await self.__scheduler.set_node_name(node_id, node_name)
            await write_string(name)

        async def comm_duplicate_nodes():  # elif command == b'duplicatenodes':
            cnt = struct.unpack('>Q', await reader.readexactly(8))[0]
            node_ids = [struct.unpack('>Q', await reader.readexactly(8))[0] for _ in range(cnt)]
            try:
                res = await self.__scheduler.duplicate_nodes(node_ids)
            except Exception as e:
                writer.write(b'\0')
            else:
                writer.write(b'\1')
                writer.write(struct.pack('>Q', len(res)))
                for old_id, new_id in res.items():
                    writer.write(struct.pack('>QQ', old_id, new_id))

        # node connection related commands
        async def comm_change_connection():  # elif command == b'changeconnection':
            connection_id, change_out, change_in, new_id_out, new_id_in = struct.unpack('>Q??QQ', await reader.readexactly(26))
            in_name, out_name = None, None
            if change_out:
                out_name = await read_string()
            else:
                new_id_out = None
            if change_in:
                in_name = await read_string()
            else:
                new_id_in = None
            await self.__scheduler.change_node_connection(connection_id, new_id_out, out_name, new_id_in, in_name)
            writer.write(b'\1')

        async def comm_add_connection():  # elif command == b'addconnection':
            id_out, id_in = struct.unpack('>QQ', await reader.readexactly(16))
            out_name = await read_string()
            in_name = await read_string()
            connection_id = await self.__scheduler.add_node_connection(id_out, out_name, id_in, in_name)
            writer.write(struct.pack('>Q', connection_id))

        async def comm_remove_connection():  # elif command == b'removeconnection':
            connection_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            await self.__scheduler.remove_node_connection(connection_id)
            writer.write(b'\1')

        # task mod
        async def comm_pause_tasks():  # elif command == b'tpauselst':  # pause tasks
            task_ids = [-1]
            numtasks, paused, task_ids[0] = struct.unpack('>Q?Q', await reader.readexactly(17))  # there will be at least 1 task, cannot be zero
            if numtasks > 1:
                task_ids += struct.unpack('>' + 'Q' * (numtasks - 1), await reader.readexactly(8 * (numtasks - 1)))
            await self.__scheduler.set_task_paused(task_ids, bool(paused))
            writer.write(b'\1')

        async def comm_pause_task_group():  # elif command == b'tpausegrp':  # pause task group
            paused = struct.unpack('>?', await reader.readexactly(1))[0]
            task_group = await read_string()
            await self.__scheduler.set_task_paused(task_group, bool(paused))
            writer.write(b'\1')

        async def comm_archive_task_group():  # elif command == b'tarchivegrp':  # archive or unarchive a task group
            group_name = await read_string()
            state = TaskGroupArchivedState(struct.unpack('>I', await reader.readexactly(4))[0])
            await self.__scheduler.set_task_group_archived(group_name, state)
            writer.write(b'\1')

        async def comm_task_cancel():  # elif command == b'tcancel':  # cancel task invocation
            task_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            await self.__scheduler.cancel_invocation_for_task(task_id)
            writer.write(b'\1')

        async def worker_task_cancel():  # workertaskcancel  # cancel invocation for worker
            worker_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            await self.__scheduler.cancel_invocation_for_worker(worker_id)
            writer.write(b'\1')

        async def comm_task_set_node():  # elif command == b'tsetnode':  # set task node
            task_id, node_id = struct.unpack('>QQ', await reader.readexactly(16))
            await self.__scheduler.force_set_node_task(task_id, node_id)
            writer.write(b'\1')

        async def comm_task_change_state():  # elif command == b'tcstate':  # change task state
            task_ids = [-1]
            numtasks, state, task_ids[0] = struct.unpack('>QIQ', await reader.readexactly(20))  # there will be at least 1 task, cannot be zero
            if numtasks > 1:
                task_ids += struct.unpack('>' + 'Q' * (numtasks - 1), await reader.readexactly(8 * (numtasks - 1)))
            await self.__scheduler.force_change_task_state(task_ids, TaskState(state))
            writer.write(b'\1')

        async def comm_task_set_name():  # elif command == b'tsetname':  # change task name
            task_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            new_name = await read_string()
            await self.__scheduler.set_task_name(task_id, new_name)
            writer.write(b'\1')

        async def comm_task_set_groups():  # elif command == b'tsetgroups':  # set task groups
            task_id, strcount = struct.unpack('>QQ', await reader.readexactly(16))
            task_groups = set()
            for _ in range(strcount):
                task_groups.add(await read_string())
            await self.__scheduler.set_task_groups(task_id, task_groups)
            writer.write(b'\1')

        async def comm_task_update_attribs():  # elif command == b'tupdateattribs':  # UPDATE task attribs, not set and override completely
            task_id, update_data_size, strcount = struct.unpack('>QQQ', await reader.readexactly(24))
            attribs_to_update = await asyncio.get_event_loop().run_in_executor(None, pickle.loads, await reader.readexactly(update_data_size))
            attribs_to_delete = set()
            for _ in range(strcount):
                attribs_to_delete.add(await read_string())
            await self.__scheduler.update_task_attributes(task_id, attribs_to_update, attribs_to_delete)
            writer.write(b'\1')

        # create new task
        async def comm_add_task():  # elif command == b'addtask':
            tasksize = struct.unpack('>Q', await reader.readexactly(8))[0]
            newtask: NewTask = NewTask.deserialize(await reader.readexactly(tasksize))
            ret: Tuple[SpawnStatus, Optional[int]] = await self.__scheduler.spawn_tasks(newtask)
            writer.write(struct.pack('>I?Q', ret[0].value, ret[1] is not None, 0 if ret[1] is None else ret[1]))

        #
        async def set_task_environment_resolver_arguments():
            task_id, data_size = struct.unpack('>QQ', await reader.readexactly(16))
            if data_size == 0:
                env_res = None
            else:
                env_res = await EnvironmentResolverArguments.deserialize_async(await reader.readexactly(data_size))
            await self.__scheduler.set_task_environment_resolver_arguments(task_id, env_res)
            writer.write(b'\1')

        #
        async def set_worker_groups():
            hwid, groups_count = struct.unpack('>QQ', await reader.readexactly(16))
            groups = []
            for i in range(groups_count):
                groups.append(await read_string())

            await self.__scheduler.set_worker_groups(hwid, groups)
            writer.write(b'\1')

        #
        #
        commands = {#'getfullstate': comm_get_full_state,
                    'get_db_uid': comm_get_db_uid,
                    'get_ui_graph_state_update_id': comm_get_ui_graph_state_update_id,
                    'get_ui_graph_state': comm_get_ui_graph_state,
                    'get_ui_task_groups': comm_get_ui_task_groups,
                    'get_ui_tasks_state': comm_get_ui_tasks_state,
                    'request_task_events': comm_request_subscribe_to_task_events,
                    'task_events_since_id': comm_request_task_events_since_id,
                    'get_ui_workers_state': comm_get_ui_workers_state,
                    'getinvocmeta': comm_get_invoc_meta,
                    'getlog': comm_get_log,
                    'getnodeinterface': comm_get_node_interface,
                    'gettaskattribs': comm_get_task_attribs,
                    'gettaskinvoc': comm_get_task_invocation,
                    'listnodetypes': comm_list_node_types,
                    'listnodepresets': comm_list_presets,
                    'getnodepreset': comm_get_node_preset,
                    'removenode': comm_remove_node,
                    'addnode': comm_add_node,
                    'wipenode': comm_wipe_node,
                    'nodehasparam': comm_node_has_param,
                    'setnodeparam': comm_set_node_param,
                    'batchsetnodeparams': comm_set_node_params,
                    # 'setnodeparamexpression': comm_set_node_param_expression,
                    'applynodesettings': comm_apply_node_settings,
                    'savecustomnodesettings': comm_save_custom_node_settings,
                    'setsettingsdefault': comm_set_settings_default,
                    'renamenode': comm_rename_node,
                    'duplicatenodes': comm_duplicate_nodes,
                    'changeconnection': comm_change_connection,
                    'addconnection': comm_add_connection,
                    'removeconnection': comm_remove_connection,
                    'tpauselst': comm_pause_tasks,
                    'tpausegrp': comm_pause_task_group,
                    'tarchivegrp': comm_archive_task_group,
                    'tcancel': comm_task_cancel,
                    'workertaskcancel': worker_task_cancel,
                    'tsetnode': comm_task_set_node,
                    'tcstate': comm_task_change_state,
                    'tsetname': comm_task_set_name,
                    'tsetgroups': comm_task_set_groups,
                    'tupdateattribs': comm_task_update_attribs,
                    'addtask': comm_add_task,
                    'settaskenvresolverargs': set_task_environment_resolver_arguments,
                    'setworkergroups': set_worker_groups}
        #
        # connection callback
        #
        self.__logger.debug('UI connected')

        async def read_string() -> str:
            strlen = struct.unpack('>Q', await reader.readexactly(8))[0]
            if strlen == 0:
                return ''
            return (await reader.readexactly(strlen)).decode('UTF-8')

        async def write_string(s: str):
            b = s.encode('UTF-8')
            writer.write(struct.pack('>Q', len(b)))
            writer.write(b)

        try:
            proto = await asyncio.wait_for(reader.readexactly(4), timeout=self.__timeout)
            if proto != b'\0\1\0\0':
                raise NotImplementedError(f'protocol version unsupported {proto}')

            wait_stop_task = asyncio.create_task(self.__scheduler._stop_event_wait())
            while True:
                readline_task = asyncio.create_task(read_string())
                done, pending = await asyncio.wait((readline_task, wait_stop_task), return_when=asyncio.FIRST_COMPLETED)
                if wait_stop_task in done:
                    self.__logger.debug(f'scheduler is stopping: closing ui connections')
                    readline_task.cancel()
                    return
                assert readline_task.done()
                try:
                    command: str = await readline_task
                except IncompleteReadError:  # this means connection was closed
                    self.__logger.debug('UI disconnected: connection closed')
                    break
                self.__logger.debug(f'got command {command}')
                # get full nodegraph state. only brings in where is which item, no other details

                if command in commands:
                    await commands[command]()
                #
                # if conn is closed - result will be b'', but in mostl likely totally impossible case it can be unfinished command.
                # so lets just catch all
                elif reader.at_eof():
                    self.__logger.debug('UI disconnected')
                    return
                else:
                    raise NotImplementedError()

                await writer.drain()
        except ConnectionResetError as e:
            self.__logger.warning('connection was reset. UI disconnected %s', e)
        except ConnectionError as e:
            self.__logger.error('connection error. UI disconnected %s', e)
        except Exception as e:
            self.__logger.exception('unknown error. UI disconnected %s', e)
        finally:
            writer.close()
            await writer.wait_closed()
            # according to the note in the beginning of the function - now reference can be cleared
            self.__saved_references.remove(asyncio.current_task())


#
# ======== client ========
#


class UIProtocolSocketClient:
    def __init__(self, host: str, port: int, timeout=30):
        self.__address = (host, port)
        self.__connection: Optional[BufferedConnection] = None
        self.__timeout = timeout

    def initialize(self):
        if self.__connection is None:
            self.__connection = BufferedConnection(self.__address, timeout=self.__timeout)
            self.__connection.writer.write(b'\0\1\0\0')

    def close(self):
        if self.__connection is None:
            return
        self.__connection.close()

    def __enter__(self):
        self.initialize()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    # def get_full_state(self, skip_dead, skip_archived_groups) -> UiData:
    #     r, w = self.__connection.get_rw_pair()
    #     w.write_string('getfullstate')
    #     w.write(struct.pack('>??', skip_dead, skip_archived_groups))
    #     w.flush()
    #     return UiData.deserialize(r)

    def get_db_uid(self) -> int:
        """

        :return:
        """
        r, w = self.__connection.get_rw_pair()
        w.write_string('get_db_uid')
        w.flush()
        uid, = struct.unpack('>Q', r.readexactly(8))
        return uid

    def get_ui_graph_state_update_id(self) -> int:
        """
        get latest graph update event id
        """
        r, w = self.__connection.get_rw_pair()
        w.write_string('get_ui_graph_state_update_id')
        w.flush()
        uid, = struct.unpack('>Q', r.readexactly(8))
        return uid

    def get_ui_graph_state(self) -> (NodeGraphStructureData, int):
        """
        returns latest graph state and update event id corresponding to it
        that id can be used to check against get_ui_graph_state_update_id() to check if update is available
        """
        r, w = self.__connection.get_rw_pair()
        w.write_string('get_ui_graph_state')
        w.flush()
        update_id, = struct.unpack('>Q', r.readexactly(8))
        data = NodeGraphStructureData.deserialize(r)
        return data, update_id

    def get_ui_task_groups(self, skip_archived_groups) -> TaskGroupBatchData:
        r, w = self.__connection.get_rw_pair()
        w.write_string('get_ui_task_groups')
        w.write(struct.pack('>?', skip_archived_groups))
        w.flush()
        data = TaskGroupBatchData.deserialize(r)
        return data

    def get_ui_tasks_state(self, groups: Iterable[str], include_dead: bool) -> TaskBatchData:
        r, w = self.__connection.get_rw_pair()
        if not isinstance(groups, (list, tuple)):
            groups = list(groups)
        w.write_string('get_ui_tasks_state')
        w.write(struct.pack('>?Q', include_dead, len(groups)))
        for group in groups:
            w.write_string(group)
        w.flush()
        data = TaskBatchData.deserialize(r)
        return data

    # task subscription
    def request_subscribe_to_task_events(self, groups: Iterable[str], include_dead: bool, request_for_seconds: float = 10.0) -> List[TaskEvent]:
        r, w = self.__connection.get_rw_pair()
        if not isinstance(groups, (list, tuple)):
            groups = list(groups)
        w.write_string('request_task_events')
        w.write(struct.pack('>?Qd', include_dead, len(groups), request_for_seconds))
        for group in groups:
            w.write_string(group)
        w.flush()
        num_events, = struct.unpack('>Q', r.readexactly(8))
        events = []
        for i in range(num_events):
            events.append(TaskEvent.deserialize(r))
        return events

    def request_task_events_since_id(self, groups: Iterable[str], include_dead: bool, last_known_event_id: int) -> Optional[List[TaskEvent]]:
        """
        returns None if there is no valid subscription with given parameters,
        returns list of task events otherwise
        """
        r, w = self.__connection.get_rw_pair()
        if not isinstance(groups, (list, tuple)):
            groups = list(groups)
        w.write_string('task_events_since_id')
        w.write(struct.pack('>?QQ', include_dead, len(groups), last_known_event_id))
        for group in groups:
            w.write_string(group)
        w.flush()
        good, = struct.unpack('>?', r.readexactly(1))
        if not good:
            return None
        num_events, = struct.unpack('>Q', r.readexactly(8))
        events = []
        for i in range(num_events):
            events.append(TaskEvent.deserialize(r))
        return events
    #

    def get_ui_workers_state(self) -> WorkerBatchData:
        r, w = self.__connection.get_rw_pair()
        w.write_string('get_ui_workers_state')
        w.flush()
        data = WorkerBatchData.deserialize(r)
        return data

    def get_invoc_meta(self, task_id) -> Dict[int, List[IncompleteInvocationLogData]]:
        r, w = self.__connection.get_rw_pair()
        w.write_string('getinvocmeta')
        w.write(struct.pack('>Q', task_id))
        w.flush()

        ret: Dict[int, List[IncompleteInvocationLogData]] = {}
        node_count, = struct.unpack('>Q', r.readexactly(8))
        for _ in range(node_count):
            node_id, invoc_count = struct.unpack('>QQ', r.readexactly(16))
            for _ in range(invoc_count):
                log = IncompleteInvocationLogData.deserialize(r)
                ret.setdefault(node_id, []).append(log)

        return ret

    def get_log(self, invocation_id) -> Optional[InvocationLogData]:
        r, w = self.__connection.get_rw_pair()
        w.write_string('getlog')
        w.write(struct.pack('>Q', invocation_id))
        w.flush()
        has_data, = struct.unpack('>?', r.readexactly(1))
        log = None
        if has_data:
            log = InvocationLogData.deserialize(r)
        return log

    def get_log_all(self, task_id, node_id):
        raise NotImplementedError()

    def get_node_interface(self, node_id) -> NodeUi:
        r, w = self.__connection.get_rw_pair()
        w.write_string('getnodeinterface')
        w.write(struct.pack('>Q', node_id))
        w.flush()
        rcvsize = struct.unpack('>I', r.readexactly(4))[0]
        nodeui: NodeUi = pickle.loads(r.readexactly(rcvsize))
        return nodeui

    def get_task_attribs(self, task_id) -> Tuple[Dict[str, Any], Optional[EnvironmentResolverArguments]]:
        r, w = self.__connection.get_rw_pair()
        w.write_string('gettaskattribs')
        w.write(struct.pack('>Q', task_id))
        w.flush()
        rcvsize = struct.unpack('>Q', r.readexactly(8))[0]
        attribs = _deserialize_json_dict(r.readexactly(rcvsize))
        rcvsize = struct.unpack('>Q', r.readexactly(8))[0]
        env_attrs = None
        if rcvsize > 0:
            env_attrs = EnvironmentResolverArguments.deserialize(r.readexactly(rcvsize))
        return attribs, env_attrs

    def get_task_invocation(self, task_id) -> InvocationJob:
        r, w = self.__connection.get_rw_pair()
        w.write_string('gettaskinvoc')
        w.write(struct.pack('>Q', task_id))
        w.flush()
        rcvsize = struct.unpack('>Q', r.readexactly(8))[0]
        if rcvsize == 0:
            invoc = InvocationJob([])
        else:
            invoc = InvocationJob.deserialize(r.readexactly(rcvsize))
        return invoc

    def list_node_types(self) -> Dict[str, NodeTypeMetadata]:
        r, w = self.__connection.get_rw_pair()
        metas: List[NodeTypeMetadata] = []
        w.write_string('listnodetypes')
        w.flush()
        elemcount, = struct.unpack('>Q', r.readexactly(8))
        for i in range(elemcount):
            btlen, = struct.unpack('>Q', r.readexactly(8))
            metas.append(pickle.loads(r.readexactly(btlen)))
        nodetypes = {n.type_name: n for n in metas}
        return nodetypes

    def list_presets(self) -> Dict[str, Dict[str, NodeSnippetDataPlaceholder]]:
        r, w = self.__connection.get_rw_pair()
        w.write_string('listnodepresets')
        w.flush()
        btlen, = struct.unpack('>Q', r.readexactly(8))
        presets = pickle.loads(r.readexactly(btlen))
        return presets

    def get_node_preset(self, package_name, preset_name) -> Optional[NodeSnippetData]:
        r, w = self.__connection.get_rw_pair()
        w.write_string('getnodepreset')
        w.write_string(package_name)
        w.write_string(preset_name)
        w.flush()
        good, = struct.unpack('>?', r.readexactly(1))
        if not good:
            return None
        btlen = struct.unpack('>Q', r.readexactly(8))[0]
        snippet: NodeSnippetData = NodeSnippetData.deserialize(r.readexactly(btlen))
        return snippet

    def remove_node(self, node_id: int) -> bool:
        r, w = self.__connection.get_rw_pair()
        w.write_string('removenode')
        w.write(struct.pack('>Q', node_id))
        w.flush()
        return r.readexactly(1) == b'\1'

    def add_node(self, node_type: str, node_name: str) -> int:
        r, w = self.__connection.get_rw_pair()
        w.write_string('addnode')
        w.write_string(node_type)
        w.write_string(node_name)
        w.flush()
        node_id, = struct.unpack('>Q', r.readexactly(8))
        return node_id

    def wipe_node(self, node_id: int):
        r, w = self.__connection.get_rw_pair()
        w.write_string('wipenode')
        w.write(struct.pack('>Q', node_id))
        w.flush()
        assert r.readexactly(1) == b'\1'

    def node_has_param(self, node_id: int, param_name: str) -> bool:
        r, w = self.__connection.get_rw_pair()
        w.write_string('nodehasparam')
        w.write(struct.pack('>Q', node_id))
        w.write_string(param_name)
        w.flush()
        good = r.readexactly(1) == b'\1'
        return good

    def set_node_param(self, node_id: int, param: Parameter) -> Union[int, float, str, bool, None]:
        """

        :param node_id:
        :param param:
        :return: newly set parameter value, or None if failed
        """
        r, w = self.__connection.get_rw_pair()
        param_type = param.type()
        param_value = param.unexpanded_value()
        w.write_string('setnodeparam')
        w.write(struct.pack('>QI?', node_id, param_type.value, param.has_expression()))
        w.write_string(param.name())
        if param.has_expression():
            w.write_string(param.expression())
        newval = None
        if param_type == NodeParameterType.FLOAT:
            w.write(struct.pack('>d', param_value))
            w.flush()
            if r.readexactly(1) == b'\1':
                newval, = struct.unpack('>d', r.readexactly(8))
        elif param_type == NodeParameterType.INT:
            w.write(struct.pack('>q', param_value))
            w.flush()
            if r.readexactly(1) == b'\1':
                newval, = struct.unpack('>q', r.readexactly(8))
        elif param_type == NodeParameterType.BOOL:
            w.write(struct.pack('>?', param_value))
            w.flush()
            if r.readexactly(1) == b'\1':
                newval, = struct.unpack('>?', r.readexactly(1))
        elif param_type == NodeParameterType.STRING:
            w.write_string(param_value)
            w.flush()
            if r.readexactly(1) == b'\1':
                newval = r.read_string()
        else:
            raise NotImplementedError()
        return newval

    def set_node_params(self, node_id: int, params: Iterable[Parameter], want_result: bool = False) -> Optional[Dict[str, Union[int, float, str, bool, None]]]:
        r, w = self.__connection.get_rw_pair()
        params = list(params)
        w.write_string('batchsetnodeparams')
        w.write(struct.pack('>QQ?', node_id, len(params), want_result))
        for param in params:
            param_type = param.type()
            w.write_string(param.name())
            value_is_expression = param.has_expression()
            w.write(struct.pack('>I?', param_type.value, value_is_expression))
            if value_is_expression:
                w.write_string(param.expression())
            else:
                param_value = param.unexpanded_value()
                if param_type == NodeParameterType.FLOAT:
                    w.write(struct.pack('>d', param_value))
                elif param_type == NodeParameterType.INT:
                    w.write(struct.pack('>q', param_value))
                elif param_type == NodeParameterType.BOOL:
                    w.write(struct.pack('>?', param_value))
                elif param_type == NodeParameterType.STRING:
                    w.write_string(param_value)
                else:
                    raise NotImplementedError()
        w.flush()
        if not want_result:
            assert r.readexactly(1) == b'\1'  # not expect result, just ack
            return
        result = {}
        for param in params:
            if r.readexactly(1) == b'\0':
                result[param.name()] = None
                continue
            param_type = param.type()
            if param_type == NodeParameterType.FLOAT:
                val, = struct.unpack('>d', r.readexactly(8))
            elif param_type == NodeParameterType.INT:
                val, = struct.unpack('>q', r.readexactly(8))
            elif param_type == NodeParameterType.BOOL:
                val, = struct.unpack('>?', r.readexactly(1))
            elif param_type == NodeParameterType.STRING:
                val = r.read_string()
            else:
                raise NotImplementedError()
            result[param.name()] = val
        return result

    def set_node_param_expression(self, node_id: int, param_name: str, expression: Optional[str]) -> bool:
        r, w = self.__connection.get_rw_pair()
        set_or_unset = expression is not None
        w.write_string('setnodeparamexpression')
        w.write(struct.pack('>Q?', node_id, set_or_unset))
        w.write_string(param_name)
        if set_or_unset:
            w.write_string(expression)
        w.flush()
        return r.readexactly(1) == b'\1'

    def apply_node_settings(self, node_id: int, settings_name: str) -> bool:
        r, w = self.__connection.get_rw_pair()
        w.write_string('applynodesettings')
        w.write(struct.pack('>Q', node_id))
        w.write_string(settings_name)
        w.flush()
        return r.readexactly(1) == b'\1'

    def save_custom_node_settings(self, node_type_name: str, settings_name: str, settings: Dict[str, Any]) -> bool:
        r, w = self.__connection.get_rw_pair()
        w.write_string('savecustomnodesettings')
        w.write_string(node_type_name)
        w.write_string(settings_name)
        settings_data = pickle.dumps(settings)
        w.write(struct.pack('>Q', len(settings_data)))
        w.write(settings_data)
        w.flush()
        return r.readexactly(1) == b'\1'

    def set_settings_default(self, node_type_name: str, settings_name: Optional[str]):
        r, w = self.__connection.get_rw_pair()
        w.write_string('setsettingsdefault')
        w.write_string(node_type_name)
        w.write(struct.pack('>?', settings_name is not None))
        if settings_name is not None:
            w.write_string(settings_name)
        w.flush()
        return r.readexactly(1) == b'\1'

    def rename_node(self, node_id: int, node_name: str) -> str:
        r, w = self.__connection.get_rw_pair()
        w.write_string('renamenode')
        w.write(struct.pack('>Q', node_id))
        w.write_string(node_name)
        w.flush()
        return r.read_string()

    def duplicate_nodes(self, node_ids: List[int]) -> Optional[Dict[int, int]]:
        """

        :param node_ids:
        :return:  mapping of old node ids to new node ids
        """
        r, w = self.__connection.get_rw_pair()
        w.write_string('duplicatenodes')
        w.write(struct.pack('>Q', len(node_ids)))
        for node_id in node_ids:
            w.write(struct.pack('>Q', node_id))
        w.flush()
        result = r.readexactly(1)
        if result == b'\0':
            return None
        cnt, = struct.unpack('>Q', r.readexactly(8))
        ret = {}
        for i in range(cnt):
            old_id, new_id = struct.unpack('>QQ', r.readexactly(16))
            assert old_id in node_ids
            ret[old_id] = new_id
        return ret

    def change_connection_by_id(self, connection_id: int, outnode_id: Optional[int], out_name: Optional[str], innode_id: Optional[int], in_name: Optional[str]):
        r, w = self.__connection.get_rw_pair()
        w.write_string('changeconnection')
        w.write(struct.pack('>Q??QQ', connection_id, outnode_id is not None, innode_id is not None, outnode_id or 0, innode_id or 0))
        if outnode_id is not None:
            w.write_string(out_name)
        if innode_id is not None:
            w.write_string(in_name)
        w.flush()
        assert r.readexactly(1) == b'\1'

    def add_connection(self, outnode_id: int, out_name: str, innode_id: int, in_name: str) -> int:
        """

        :param outnode_id:
        :param out_name:
        :param innode_id:
        :param in_name:
        :return: new or existing connection id
        """
        r, w = self.__connection.get_rw_pair()
        w.write_string('addconnection')
        w.write(struct.pack('>QQ', outnode_id, innode_id))
        w.write_string(out_name)
        w.write_string(in_name)
        w.flush()
        new_or_existing_id, = struct.unpack('>Q', r.readexactly(8))
        return new_or_existing_id

    def remove_connection_by_id(self, connection_id: int):
        r, w = self.__connection.get_rw_pair()
        w.write_string('removeconnection')
        w.write(struct.pack('>Q', connection_id))
        w.flush()
        assert r.readexactly(1) == b'\1'

    def pause_tasks(self, task_ids: Iterable[int], paused: bool):
        r, w = self.__connection.get_rw_pair()
        task_ids = list(task_ids)
        numtasks = len(task_ids)
        if numtasks == 0:
            return
        w.write_string('tpauselst')
        w.write(struct.pack('>Q?Q', numtasks, paused, task_ids[0]))
        if numtasks > 1:
            w.write(struct.pack('>' + 'Q' * (numtasks - 1), *task_ids[1:]))
        w.flush()
        assert r.readexactly(1) == b'\1'

    def pause_task_group(self, task_group_name, paused):
        r, w = self.__connection.get_rw_pair()
        w.write_string('tpausegrp')
        w.write(struct.pack('>?', paused))
        w.write_string(task_group_name)
        w.flush()
        assert r.readexactly(1) == b'\1'

    def archive_task_group(self, archived_state: TaskGroupArchivedState, task_group_name: str):
        r, w = self.__connection.get_rw_pair()
        w.write_string('tarchivegrp')
        w.write_string(task_group_name)
        w.write(struct.pack('>I', archived_state.value))
        w.flush()
        assert r.readexactly(1) == b'\1'

    def cancel_invocation_for_task(self, task_id: int):
        r, w = self.__connection.get_rw_pair()
        w.write_string('tcancel')
        w.write(struct.pack('>Q', task_id))
        w.flush()
        assert r.readexactly(1) == b'\1'

    def cancel_invocation_for_worker(self, worker_id: int):
        r, w = self.__connection.get_rw_pair()
        w.write_string('workertaskcancel')
        w.write(struct.pack('>Q', worker_id))
        w.flush()
        assert r.readexactly(1) == b'\1'

    def set_node_for_task(self, task_id: int, node_id: int):
        r, w = self.__connection.get_rw_pair()
        w.write_string('tsetnode')
        w.write(struct.pack('>QQ', task_id, node_id))
        w.flush()
        assert r.readexactly(1) == b'\1'

    def change_tasks_state(self, task_ids: Iterable[int], state: TaskState):
        r, w = self.__connection.get_rw_pair()
        task_ids = list(task_ids)
        numtasks = len(task_ids)
        if numtasks == 0:
            return
        w.write_string('tcstate')
        w.write(struct.pack('>QIQ', numtasks, state.value, task_ids[0]))
        if numtasks > 1:
            w.write(struct.pack('>' + 'Q' * (numtasks - 1), *task_ids[1:]))
        w.flush()
        assert r.readexactly(1) == b'\1'

    def set_task_name(self, task_id: int, new_name: str):
        r, w = self.__connection.get_rw_pair()
        w.write_string('tsetname')
        w.write(struct.pack('>Q', task_id))
        w.write_string(new_name)
        w.flush()
        assert r.readexactly(1) == b'\1'

    def set_task_groups(self, task_id: int, task_groups: Iterable[str]):
        r, w = self.__connection.get_rw_pair()
        task_groups = list(task_groups)
        w.write_string('tsetgroups')
        w.write(struct.pack('>QQ', task_id, len(task_groups)))
        for group in task_groups:
            w.write_string(group)
        w.flush()
        assert r.readexactly(1) == b'\1'

    def update_task_attributes(self, task_id, attribs_to_set: Dict[str, Any], attribs_to_delete: Iterable[str]):
        r, w = self.__connection.get_rw_pair()
        attribs_to_delete = set(attribs_to_delete)
        w.write_string('tupdateattribs')
        data_bytes = pickle.dumps(attribs_to_set)
        w.write(struct.pack('>QQQ', task_id, len(data_bytes), len(attribs_to_delete)))
        w.write(data_bytes)
        for attr in attribs_to_delete:
            w.write_string(attr)
        w.flush()
        assert r.readexactly(1) == b'\1'

    def add_task(self, new_task: NewTask) -> Optional[int]:
        """

        :param new_task:
        :return:  new task id if succeeded, else - None
        """
        r, w = self.__connection.get_rw_pair()
        data = new_task.serialize()
        w.write_string('addtask')
        w.write(struct.pack('>Q', len(data)))
        w.write(data)
        w.flush()
        spawn_status_value, good, new_id = struct.unpack('>I?Q', r.readexactly(13))  # reply that we don't care about for now
        if SpawnStatus(spawn_status_value) == SpawnStatus.FAILED or not good:
            return None
        return new_id

    def set_task_environment_resolver_arguments(self, task_id: int, env_args: Optional[EnvironmentResolverArguments]):
        r, w = self.__connection.get_rw_pair()
        w.write_string('settaskenvresolverargs')
        w.write(struct.pack('>Q', task_id))
        if env_args is None:
            w.write(struct.pack('>Q', 0))
        else:
            data = env_args.serialize()
            w.write(struct.pack('>Q', len(data)))
            w.write(data)
        w.flush()
        assert r.readexactly(1) == b'\1'

    def set_worker_groups(self, worker_hwid: int, groups: Iterable[str]):
        groups = list(groups)
        r, w = self.__connection.get_rw_pair()
        w.write_string('setworkergroups')
        w.write(struct.pack('>QQ', worker_hwid, len(groups)))
        for group in groups:
            w.write_string(group)
        w.flush()
        assert r.readexactly(1) == b'\1'




def scheduler_port() -> int:
    return 1384


def scheduler_message_port() -> int:
    return 1386


def ui_port() -> int:
    return 1385


def broadcast_port() -> int:
    return 34305


def worker_pool_port() -> int:
    return 1387


def worker_start_port() -> int:
    return 6969


def message_proxy_port() -> int:
    return 22182
